/****************************** Module Header *******************************
*
* Module Name: file.e
*
* Copyright (c) Netlabs EPM Distribution Project 2004
*
* $Id: file.e 4790 2021-11-21 10:30:24Z aschn $
*
* ===========================================================================
*
* This file is part of the Netlabs EPM Distribution package and is free
* software.  You can redistribute it and/or modify it under the terms of the
* GNU General Public License as published by the Free Software
* Foundation, in version 2 as it comes in the "COPYING" file of the
* Netlabs EPM Distribution.  This library is distributed in the hope that it
* will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
****************************************************************************/

; File commands and procedures.
; Moved from STDCMDS.E, STDPROCS.E.
; See also: SLNOHOST.E/SAVELOAD.E/E3EMUL.E.

compile if not defined( SMALL)  -- If compiled separately
   EA_comment 'This defines macros for file operations.'

define
   INCLUDING_FILE = 'FILE.E'

   include 'stdconst.e'
const
   tryinclude 'mycnf.e'
const
 compile if not defined( NLS_LANGUAGE)
   NLS_LANGUAGE = 'ENGLISH'
 compile endif
   include NLS_LANGUAGE'.e'
compile endif

; ---------------------------------------------------------------------------
const
-- Lets you quit temporary files regardless of the state of the .modify bit.
-- Temporary files are assumed to be any file where the first character of the
-- .filename is a period.  If set to 1, you won't get the "Throw away changes?"
-- prompt when trying to quit one of these files.
compile if not defined( TRASH_TEMP_FILES)
   TRASH_TEMP_FILES = 0
compile endif

; ---------------------------------------------------------------------------
; Called after MB2 dclick on titlebar and Enter.
; Fixed to work with enhanced titletext.
; ProcessName triggers a defselect event.
defc ProcessName

   -- Find (old) .filename in (old) .titletext
   Sep = GetFieldSep()
   Rest = upcase( .titletext) Sep
   Filename = upcase( .filename)
   s = 0   -- amount of Seps
   do while rest <> ''
      parse value Rest with Next (Sep) Rest
      if strip( Next) = strip( Filename) then
         leave
      endif
      s = s + 1
   enddo

   -- Remove the other fields from arg( 1), parsed at Sep
   Rest = arg( 1) Sep
   do n = 1 to s + 1
      parse value Rest with Next (Sep) Rest
      Next = strip( Next)
      Rest = strip( Rest)
      if Rest = '' then
         leave
      endif
   enddo
   NewFilename = strip( Next)
   if NewFilename = '' then
      NewFilename = arg( 1)
   endif

   if NewFilename <> '' & NewFilename <> .filename then
      'Name' NewFilename
   endif

; ---------------------------------------------------------------------------
defc Name, n
   do once = 1 to 1
      NewFilename = arg( 1)
      if NewFilename = '' then
         -- Name without args supplies current name
         'CommandLine Name '.filename
      else
         -- Unlock if locked
         if .lockhandle then
            'Unlock'
            if rc then
               leave
            endif
         endif

         -- Rename current file
         OldFilename = .filename
         AutosaveFilename = MakeTempName()
         call NameFile( NewFilename)
         if rc then
            leave
         endif
         if OldFilename = .filename then
            leave
         endif

         -- Advance .modify
         .modify = .modify + 1

         -- Handle longname
         NewName = GetFileSpec( 'N', NewFilename)
         if MakeFilename( NewName) <> NewName then
            'AddEa .LONGNAME 'NewName
         else
            if Get_EAT_ASCII_Value( '.LONGNAME') <> '' then
               call Delete_Ea( '.LONGNAME')
            endif
         endif

         -- Remove .readonly field if original file was .readonly and new
         -- file doesn't exist. (If file exists, then the attrib is
         -- re-determined correctly.)
         if .readonly then
            if not Exist( .filename) then
               .readonly = 0
            endif
            rc = 0
         endif

         -- Rename the autosave file
         call MoveFile( AutosaveFilename, MakeTempName())
         rc = 0

         -- Redetermine mode
         'ResetMode'
      endif
   enddo

; ---------------------------------------------------------------------------
defc Rename
   Filename = .filename
   if Filename = GetUnnamedFilename() then
      Filename = ''
   endif

   Title = RENAME__MSG
   Text = RENAME_PROMPT__MSG 'Current directory:' directory()
   Buttons = ''
   Entry = Filename
   parse value EntryBox( Title,
                         Buttons,
                         Entry,
                         0,
                         240,
                         atoi( 1) || atoi( 0) || atol( 0) ||
                         Text) with Button 2 NewName \0

   NewName = strip( NewName)
   if Button = \1 & Newname <> '' then
      'Name' Newname
   endif

; ---------------------------------------------------------------------------
; Handles special NEPMD dirs: don't overwrite files of the NETLABS, EPMBBS or
; NETLABS src tree
defproc SaveProcessNetlabsFile( Filename)
   rc = 0
   do once = 1 to 1

      RootDir = NepmdScanEnv( 'NEPMD_ROOTDIR')
      rc = 0  -- Ignore error here
      if RootDir = '' then
         'SayError Environment var NEPMD_ROOTDIR not set.'
         leave
      endif
      UserDir = NepmdScanEnv( 'NEPMD_USERDIR')
      rc = 0  -- Ignore error here
      if UserDir = '' then
         'SayError Environment var NEPMD_USERDIR not set.'
         leave
      endif
      NepmdDevDir = NepmdScanEnv( 'NEPMD_DEVDIR')
      rc = 0  -- Ignore error here

      NetlabsDir = RootDir'\netlabs'
      EpmbbsDir  = RootDir'\epmbbs'
      if NepmdDevDir then
         NetlabsDevDir = NepmdDevDir'\src\netlabs'
      else
         NetlabsDevDir = ''
      endif

      UpName = upcase( Filename)
      if abbrev( UpName, upcase( NetlabsDir)) then
         Type = 'NETLABS'
         NewFilename = UserDir''substr( Filename, length( NetlabsDir) + 1)
      elseif abbrev( UpName, upcase( EpmbbsDir)) then
         Type = 'EPMBBS'
         NewFilename = UserDir''substr( Filename, length( EpmbbsDir) + 1)
      elseif abbrev( UpName, upcase( NetlabsDevDir)) & NetlabsDevDir then
         if rightstr( UpName, 4) = '.TMP' then
            leave
         endif
         Type = 'NEPMD source'
         NewFilename = UserDir''substr( Filename, length( NetlabsDevDir) + 1)
      else
         leave
      endif

      Title = 'Save: change to user tree'
      Text = Filename\n\n                                              ||
             'You''re about to overwrite a file of the 'Type       ||
             ' tree. Better use the user tree for your own'        ||
             ' files.'\n\n                                         ||
             'Do you want to save it to the user tree?'
      rcx = WinMessageBox( Title, Text,
                           MB_YESNOCANCEL + MB_QUERY + MB_DEFBUTTON1 + MB_MOVEABLE)

      if rcx = MBID_NO then
         rc = 0
      elseif rcx = MBID_YES then
         Filename = NewFilename
         lp = lastpos( '\', Filename)
         Newdir = leftstr( Filename, lp - 1)
         -- Create tree, if not existing
         call MakeTree( NewDir)

         NewFilename = SaveAsCheckExist( Filename)
         if rc < 0 then
            leave
         elseif rc = 0 then
            -- nop
         elseif rc = 1 then
            Filename = NewFilename
            rc = 0
         endif
      else
         rc = -5  -- sayerror( 'Access denied')
         leave
      endif
   enddo
   return Filename

; ---------------------------------------------------------------------------
; Syntax: QuitFile Filename
defc QuitFile
   Filename = StripquotesFileSpec( strip( arg( 1)))
   NumQuit = 0

   do once = 1 to 1
      if Filename = '' then
         leave
      endif

      getfileid StartFid
      'VSyncCursor'
      do f = 1 to filesinring( 1)  -- Provide an upper limit; prevent looping forever
         getfileid CurFid
         getfileid TestFid, Filename
         if TestFid = '' then
            leave
         elseif CurFid = TestFid then
            'xcom quit'
            NumQuit = NumQuit + 1
         endif
         nextfile
      enddo
      pActivateFile( StartFid)
   enddo

; ---------------------------------------------------------------------------
; This can be executed after a file with Filename is created. The file to
; quit must be on top.
; This is used to keep virtual filenames unique.
defc QuitSameFile
   Filename = .filename
   if upcase( leftstr( Filename, 5)) = '.TREE' then
      -- Strip appended SortOption
      ParseTreeTitleFilename( Filename, Filename, SortOption)
   endif

   NumQuit = 0

   getfileid StartFid
   'VSyncCursor'
   do f = 1 to filesinring( 1)  -- Provide an upper limit; prevent looping forever
      nextfile
      getfileid CurFid
      if CurFid = StartFid then
         leave
      endif
      NextFilename = .filename
      if upcase( leftstr( NextFilename, 5)) = '.TREE' then
         -- Strip appended SortOption
         ParseTreeTitleFilename( NextFilename, NextFilename, NextSortOption)
      endif
      if NextFilename = Filename then
         'xcom quit'
         NumQuit = NumQuit + 1
      endif
   enddo
   pActivateFile( StartFid)
   if NumQuit then
      -- This is also required when 'xcom quit' is replaced by 'quit'.
      -- It doesn't seem to work here automatically, so execute it here
      -- again:
      'RingSetFileNumber'
   endif

; ---------------------------------------------------------------------------
; Saves current file.
; If a filename is specified, the current file is saved to the specified
; filename and the current file isn't changed. To achieve that, Name should
; be executed before.
; If, for a temp. file, no filename is specified, the Save-as dialog is
; opened to enter a filename.
; Instead of a filename a longname may be specified. The longname is then
; converted to a filename and after save, the longname is added, if
; different from filename.
; If a <filename> equal to current file is specified, the check for an
; existing file to overwrite is bypassed.
; Syntax:
;    Save                               save
;    Save (for a tempfile)              open Save-as dialog
;    Save <filename>                    save, keep old filename loaded
;    Name <filename> followed by Save   save, change to new filename
;    call Save( <filename>, 1)          used by SaveAs
defc Save, s
   Filename = StripQuotesFileSpec( arg( 1))
   call Save( Filename)

defproc Save
   universal save_with_tabs
   universal default_save_options
   universal unnamedfilename

   do once = 1 to 1

      SpecifiedName = arg( 1)
      Name = SpecifiedName
      call Parse_Leading_Options( Name, Options)  -- gets and sets Name and Options
      Options = default_save_options Options

      fCalledBySaveAs = (arg( 2) = 1)  -- optional 2nd arg to handle call specially

      -- Workaround: This avoids unwanted window scrolling.
      'VSyncCursor'

      fTempFile    = (leftstr( .filename, 1) = '.')
      PrevMode     = GetMode()
      PrevModeSrc  = GetFileAVar( 'mode.src')
      DestLongname = ''

      -- Set name if not already or resolve it if specified
      fNameChanged = 0
      if SpecifiedName = '' then
         ----- No name specified -----
         Name = .filename

         -- Open file dialog (Save as) if filename starts with a dot, e.g. .Untitled
         if fTempFile then
            ----- Temp. file -----
            call SaveAs_Dlg( Name, Type)  -- gets and sets Name (and sets Type)
            if rc then
               return
            endif
            SpecifiedName = Name
            'Name' EnquoteFileSpec( Name)
            if not rc then
               Name = .filename
               fNameChanged = 1
            endif
         endif

      else
         ----- Name specified -----
         call Parse_Filename( Name, .filename)  -- gets .filename and sets Name

         ----- Convert longname to filename -----
         PathOfName = GetFileSpec( 'DP', Name)
         NameOfName = GetFileSpec( 'N', Name)
         Filename = PathOfName''MakeFilename( NameOfName, Name)
         -- Unique filename?
         if Filename <> Name then
            DestLongname = GetFileSpec( 'N', Name)
            Name = Filename
         endif

         ----- Query unique longname if name has changed -----
         do once2 = 1 to 1
            if DestLongname then
               leave
            endif
            fSameName = (Name = .filename)
            if fSameName then
               leave
            endif
            CurLongname = Get_Eat_Ascii_Value( '.LONGNAME')
            if CurLongname = '' then
               leave
            endif
            DestLongname = MakeUniqueLongname( Name, CurLongname)
         enddo
      endif

      ----- Netlabs file -----
      -- Don't overwrite files of the NETLABS, EPMBBS or NEPMD_DEVDIR tree
      if not fTempFile & not fCalledBySaveAs then
         ChangedName = SaveProcessNetlabsFile( Name)
         if rc then
            leave
         endif
         if upcase( ChangedName) <> upcase( Name) then
            .filename = ChangedName
            Name = ChangedName
            SpecifiedName = ChangedName
            fNameChanged = 1
         endif
      endif

      ----- Readonly -----
      -- Check for .readonly field
      if SpecifiedName = '' & (browse() | .readonly) then
         if .readonly then
            SaveMsg = READ_ONLY__MSG
         else
            SaveMsg = BROWSE_IS__MSG ON__MSG
         endif
         rc = -5  -- Access denied
         leave
      endif
      -- Check if readonly file attrib is set although .readonly field is not
      Attr = NepmdQueryPathInfo( Name, 'ATTR')  -- returns e.g. 'ADSHR' or '-----'
      if substr( Attr, 5, 1) = 'R' then
         SaveMsg = READ_ONLY__MSG
         rc = -5  -- Access denied
         leave
      endif

      ----- Unlock -----
      -- Try to unlock file if it is locked (only successful if locked by the current EPM window)
      if .lockhandle then
         'Unlock'
      endif

      ----- CONFIGSYS and pre-save hooks -----
      -- CONFIGSYS: Reconcatenate lines
      if PrevMode = 'CONFIGSYS' then
         'JoinPathLines'
      endif
      -- Execute pre-save hooks
      'HookExecute save'
      'HookExecuteOnce saveonce'

      ----- EAs -----
      -- Handle EAs
      if .levelofattributesupport bitand 8 then
         'SaveAttributes'
      endif
      KeyPath = '\NEPMD\User\AutoRestore\CursorPos'
      RestorePos = QueryConfigKey( KeyPath)
      if RestorePos = 1 then
         -- Write EPM.POS EA on save
         call pSave_Pos( CurPos)   -- get current value
         call Delete_Ea( 'EPM.POS')
         -- Don't save .scrollx, because it is not restored on file loading
         'Add_Ea EPM.POS' subword( CurPos, 1, 4)   -- that affects only .eaarea, EA is written on file writing
      endif

      ----- Options -----
      -- Append file save options
      FileSaveOptions = GetFileAVar( 'saveoptions')
      if FileSaveOptions then
         Options = Options FileSaveOptions
      endif
      Options = NormalizeSaveOptions( Options)

      ----- SaveFile -----
      -- Do the save
      display -2  -- suppress non-critical errors
      call SaveFile( Name, Options)
      src = rc
      display 2

      ----- Get a better Msg and rc -----
      SaveMsg = GetDetailedSaveMsg( src, Name)

      ----- Netlabs file: Reset name on error -----
      if src & fNameChanged then
         -- Revert to .Unnamed if no success and if Name comes from a Save-as dialog
         .filename = GetUnnamedFilename()
         Name = .filename
      endif

      ----- The rest is only for success -----
      if src then
         leave
      endif

      ----- Write DestLonganme -----
      if DestLongname then
         call NepmdWriteStringEa( Name, '.LONGNAME', DestLongname)
      endif

      ----- Mode -----
      -- Redetermine mode
      SetFileAVar( 'mode', '')
      SetFileAVar( 'mode.src', '')
      Mode    = GetMode()
      ModeSrc = GetFileAVar( 'mode.src')
      -- Save the previous mode as EA, if required
      if Mode <> PrevMode then
         if wordpos( PrevModeSrc, 'file') | PrevMode = 'SHELL' then
            -- Ensure that previous mode is set after save. The mode command
            -- handles the EPM.MODE EA itself. In this def, GetMode would return
            -- return PrevMode, not the one after save.
            'Mode 'PrevMode
         endif
      endif

      ----- History -----
      'PostMe AddToHistory SAVE' Name

      -- This has to be disabled when being called by RecompileNew ->
      -- RingCheckModify -> CheckModify -> Save.
      -- Until all these defs were changed to submit an additional option,
      -- SaveAskRecompile should better be disabled.
      --'PostMe SaveAskRecompile' Name

      ----- Post-save hooks -----
      'HookExecute postsave'
      'HookExecuteOnce postsaveonce'

      ----- ALIAS.CFG saved: rescan -----
      if upcase( Name) = upcase( Get_Env( 'NEPMD_USERDIR')'\bin\alias.cfg') then
         'PostMe ShellReadAliasFile'
      endif

      -- A refresh is required here. Otherwise an ETK MessageBox pops up
      -- with the message 'Error on file saving'. But the file was saved.
      -- This can be reproduced on closing of the EPM window while an unsaved
      -- file (not the current) is in the ring. Then a first ETK MessageBox
      -- lets one save the unsaved file, but the 2nd ETK MessageBox comes
      -- then with the wrong error message.
      -- The refresh statement doesn't suffice. The refresh command does,
      -- while the postme'd version doesn't.
      'Refresh'

   enddo  -- once

   ----- Say Msg -----
   if not IsOption( Options, 'q') then
      if SaveMsg then
         'SayError 'SaveMsg
      endif
   endif

   -- o  A defc must use "return myrc" or use "rc = myrc" to set the global
   --    var rc. When "return" is used only, rc would be set to empty.
   -- o  A defproc may return its own rcx, without overriding the global rc.
   --    This was changed to a defproc for easier arg parsing.
   rc = src
   return

; ---------------------------------------------------------------------------
; This has to be disabled when being called by RecompileNew ->
; RingCheckModify -> CheckModify -> Save.
; Until all these defs were changed to submit an additional option,
; SaveAskRecompile should better be disabled.
defc SaveAskRecompile
   do once = 1 to 1
      Name = arg( 1)
      if Name = '' then
         Name = .filename
      endif
      AskRecompileList = 'MODECFG.E STDKEYS.E CUAKEYS.E'
      Name = GetFileSpec( 'N', Name)
      len = length( 'KEYS.E')
      if rightstr( upcase( Name), len) = 'KEYS.E' & length( Name) > len then
         if not wordpos( upcase( Name), upcase( AskRecompileList)) then
            AskRecompileList = AskRecompileList upcase( Name)
         endif
      endif

      if not wordpos( upcase( Name), upcase( AskRecompileList)) then
         leave
      endif

      Title = 'Save: file requires recompile'
      Text = Name\n\n                       ||
             'needs to be recompiled.'\n\n  ||
             'Do you want to recompile it?'
      rcx = WinMessageBox( Title, Text,
                           MB_YESNO + MB_QUERY + MB_DEFBUTTON1 + MB_MOVEABLE)

      if rcx = MBID_NO then
         leave
      elseif rcx = MBID_YES then
         'RecompileNew'
      endif
   enddo

; ---------------------------------------------------------------------------
; New save options:
;    /k    keep  removes /o, /l and /a
;    /a    auto  removes /o, /l and /k, adds /o or /l from file loading
;    /bin  bin   shortcut for /k /ne /ns /nt
; For others, see: start view epmusers save, s
defproc NormalizeSaveOptions( Options)

   -- Replace option /bin with /k /ne /ns /nt in Options
   wp = ''
   do while wp <> 0
      wp = wordpos( '/bin', Options)
      if wp > 0 then
         Options = subword( Options, 1, wp - 1)' /k /ne /ns /nt 'subword( Options, wp + 1)
      endif
   enddo

   -- Replace option /u with /l /ne in Options
   wp = ''
   do while wp <> 0
      wp = wordpos( '/u', Options)
      if wp > 0 then
         Options = subword( Options, 1, wp - 1)' /l /ne 'subword( Options, wp + 1)
      endif
   enddo

   -- Normalize all options. The last option wins.
   ExcludeList = '/ns/s /nt/t /ne/e /o/l/a/k'
   NewOptions = ''
   rest = Options
   rest = translate( rest, ' ', '/')  -- replace all slashes with spaces to use the word* procs
   do while rest <> ''
      parse value rest with next rest
      next = strip( next)
      if wordpos( next, rest) <> 0 then   -- if found in rest
         iterate
      endif

      -- Find excluding options
      ExcludeWrd = ''
      do w = 1 to words( ExcludeList)
         wrd = word( ExcludeList, w)
         wrd = translate( wrd, ' ', '/')  -- replace all slashes with spaces to use the word* procs
         if wordpos( next, wrd) then
            ExcludeWrd = wrd          -- ExcludeWrd = word of ExcludeList where next belongs to, slashes are replaced
            leave
         endif
      enddo

      -- Check if rest contains options of ExcludeWrd
      Found = 0
      do x = 1 to words( ExcludeWrd)
         xopt = word( ExcludeWrd, x)
         if wordpos( xopt, rest) then
            Found = 1
            leave
         endif
      enddo
      if not Found then  -- if rest doesn't contain options of ExcludeWrd
         NewOptions = NewOptions' /'next      -- append '/'next
      endif
   enddo
   Options = strip( NewOptions)

   -- Handle auto option: Add value from file loading
   if not wordpos( '/o', Options) & not wordpos( '/l', Options) & not wordpos( '/k', Options) then
      LoadType = GetFileAVar( 'lineend')
      if upcase( LoadType) = 'UNIX' then
         Options = Options' /l'
      else
         Options = Options' /o'
      endif
   endif

   -- Remove internally unknown options /a, /k and /bin
   -- /bin is already replaced above.
   RemoveWords = '/a /k /bin'
   do w = 1 to words( RemoveWords)
      RemoveWrd = word( RemoveWords, w)
      wp = ''
      do while wp <> 0
         wp = wordpos( RemoveWrd, Options)
         if wp > 0 then
            Options = delword( Options, wp, 1)
         endif
      enddo
   enddo

   return Options

; ---------------------------------------------------------------------------
defproc GetDetailedSaveMsg( src, DestFile)
   initialrc = rc
   SaveMsg = ''
   do once = 1 to 1

      ----- Success message from SaveFile -----
      if src = 0 then

         SaveMsg = SAVED_TO__MSG DestFile
         leave

      -- Save failed: Handle filenames on FAT drives or give a better error msg
      else
         -- src = -285 then  -- ERROR_WRITING_FILE_RC  (E Toolkit)

         ----- Get message from DosOpen -----
         -- Compared to the 'save' cmd, which returns only -285 or 0,
         -- defproc Lock returns the (more usable) rc from DosOpen.
         dosopenrc = Lock( 'W')

         if dosopenrc = 0 then
            'Unlock'
         elseif dosopenrc = 206 then  -- 206 = ERROR_FILENAME_EXCED_RANGE
            src = -5
            SaveMsg = 'Source filename exceeds range'
         elseif dosopenrc = 32 then
            src = -5
            SaveMsg = ACCESS_DENIED__MSG '-' MAYBE_LOCKED__MSG  -- Maybe someone locked it.
         elseif dosopenrc = 3 then
            src = -5
            SaveMsg = 'Source path not found'
         elseif dosopenrc = 15 then
            src = -5
            SaveMsg = 'Source drive is not valid'
         elseif dosopenrc = 123 then
            src = -5
            SaveMsg = 'Source contains an illegal character'
         elseif dosopenrc = 26 then
            src = -5
            SaveMsg = 'Source is an unknown media type'
         elseif dosopenrc = 108 then
            src = -5
            SaveMsg = 'Source drive locked'
         else
            src = -5
            SaveMsg = 'DosOpen returned rc = 'dosopenrc
         endif  -- dosopenrc

         if dosopenrc <> 0 then
            leave
         endif

         ----- Get message from DosQueryPathInfo -----
         DestDir = GetFileSpec( 'DP', DestFile)
         if rightstr( DestDir, 2) <> ':\' then
            DestDir = leftstr( DestDir, length( DestDir) - 1)
         endif
         call NepmdQueryPathInfo( DestDir, 'MTIME')
         dosquerypathinforc = rc

         if dosquerypathinforc = 0 then
            -- nop
         elseif dosquerypathinforc = 3 then
            src = -5
            SaveMsg = 'Destination path not found'
         elseif dosquerypathinforc = 32 then
            src = -5
            SaveMsg = 'Destination locked'
         elseif dosquerypathinforc = 206 then
            src = -5
            SaveMsg = 'Destination path exceeds range'
         else
            src = -5
            SaveMsg = 'DosQueryPathInfo returned rc = 'dosquerypathinforc
         endif  -- dosopenrc

         if dosquerypathinforc <> 0 then
            leave
         endif

         SaveMsg = 'SafeFile returned rc = 'src

      endif
   enddo  -- once
   rc = initialrc
   return SaveMsg

; ---------------------------------------------------------------------------
; Returns a unique filename on OS/2 file systems for a passed filename.
; Syntax:
;    UniqueFilename = MakeUniqueFilename( [Filename])
defproc MakeUniqueFilename
   Filename = arg( 1)
   if Filename = '' then
      Filename = .filename
   endif

   UniqueFilename = Filename
   Path = GetFileSpec( 'DP', Filename)
   Name = GetFileSpec( 'N', Filename)
   Base = GetFileSpec( 'B', Filename)
   Ext  = GetFileSpec( 'E', Filename)

   -- Avoid empty Base
   if leftstr( Name, 1) = '.' & not pos( '.', substr( Name, 2)) then
      Base = '.'Ext
      Ext  = ''
      Base = MakeFilename( Base, Filename)
   endif

   FSys = ''
   do once = 1 to 1
      if substr( Filename, 1, 2) = '\\' then
         leave
      endif
      if substr( Filename, 2, 1) = ':' then
         FSys = QueryFileSys( substr( Filename, 1, 2))
      endif
   enddo

   nStart = 0
   if FSys = 'FAT' then
      -- Find trailing number of Base
      BaseNum = ''
      do p = length( Base) to 1 by -1
         CheckNum = rightstr( Base, p)
         if IsUInt( CheckNum) then
            BaseNum = CheckNum
            nStart = CheckNum
         else
            leave
         endif
      enddo
      BasePart = leftstr( Base, length( Base) - length( BaseNum))
      Base = BasePart
   else
      -- Find trailing '!'<number> of Base
      BasePart = Base
      lp = lastpos( '!', Base)
      if lp then
         BasePart = leftstr( Base, lp - 1)
         BaseNum  = substr( Base, lp + 1)
         if IsUInt( BaseNum) then
            nStart = BaseNum
            Base = BasePart
         endif
      endif
   endif

   UniqueBase = Base
   n = nStart
   do forever
      CheckFilename = Path''UniqueBase
      if Ext then
         CheckFilename = CheckFilename'.'Ext
      endif
      if Exist( CheckFilename) then
         n = n + 1
         if FSys = 'FAT' then
            -- Replace last digit(s) of Base with n
            UniqueBase = leftstr( UniqueBase,
                                  Min( 8, length( UniqueBase)) - length( n))''n
         else
            -- Append '!'n to Base
            UniqueBase = UniqueBase'!'n
         endif
      else
         UniqueFilename = CheckFilename
         leave
      endif
   enddo

   return UniqueFilename

; ---------------------------------------------------------------------------
; Converts a longname to a filename. Returns a name valid for OS/2 file
; systems, similar to how OS/2 would convert it. It accepts a filename or a
; drive as arg( 2) to determine the filesystem first. It defaults to the
; current filename.
; Syntax:
;    ShortName = MakeFilename( [Longname [, FileSpec]])
defproc MakeFilename
   Longname = arg( 1)
   if Longname = '' then
      Longname = NepmdQueryStringEa( .filename, '.LONGNAME')
      if Longname = '' then
         LongName = GetFileSpec( 'N', .filename)
      endif
   endif

   FileSpec = arg( 2)
   FSys = ''
   do once = 1 to 1
      if FileSpec = '' then
         FileSpec = .filename
      endif
      if substr( FileSpec, 1, 2) = '\\' then
         leave
      endif
      if substr( FileSpec, 2, 1) = ':' then
         FSys = QueryFileSys( substr( FileSpec, 1, 2))
      endif
   enddo

   -- Invalid chars are converted to an exclamation mark
   InvalidChars = ''
   InvalidChars = InvalidChars\0\1\2\3\4\5\6\7\8\9\10\11\12\13\14\15
   InvalidChars = InvalidChars\16\17\18\19\20\21\22\23\24\25\26\27\28\29\30\31
   InvalidChars = InvalidChars'"/\:|<>*?'
   if FSys = 'FAT' then
      InvalidChars = InvalidChars'+,[]=;'
   endif
   ReplaceChars = copies( '!', length( InvalidChars))

   if Fsys = 'FAT' then

      -- Space and dot are converted to an underscore
      InvalidChars = InvalidChars' .'
      ReplaceChars = ReplaceChars'__'

      -- Handle leading dot specially to avoid empty Base
      if leftstr( Longname, 1) == '.' then
         Next = '_'substr( Longname, 2)
      else
         Next = Longname
      endif

      -- Parse into Base and Ext (Ext excludes the dot)
      Base = GetFileSpec( 'B', Next)
      Ext  = GetFileSpec( 'E', Next)

      -- Maybe shorten Base and Ext to 8.3
      if length( Base) > 8 then
         Base = leftstr( Base, 8)
      endif
      if length( Ext) > 3 then
         Ext  = leftstr( Ext, 3)
      endif

      -- Convert InvalidChars of Base and Ext separately
      Base = translate( Base, ReplaceChars, InvalidChars)
      Ext  = translate( Ext,  ReplaceChars, InvalidChars)

      if Ext <> '' then
         ShortName = Base'.'Ext
      else
         ShortName = Base
      endif

   else

      -- Convert InvalidChars of Longname
      ShortName = translate( Longname, ReplaceChars, InvalidChars)

   endif

   -- Without path
   return ShortName

; ---------------------------------------------------------------------------
; Enquotes FileSpec if required.
defproc EnquoteFileSpec( FileSpec)
   if (not (leftstr( FileSpec, 1) = '"' & rightstr( FileSpec, 1) = '"')) &
      verify( FileSpec, " &|<>',", 'M') then
      FileSpec = '"'FileSpec'"'
   endif

   return FileSpec

; ---------------------------------------------------------------------------
; Removes double quotes from enquoted strings.
defproc StripQuotesFileSpec( FileSpec)
   if leftstr( FileSpec, 1) = '"' & rightstr( FileSpec, 1) = '"' &
      length( FileSpec) > 1 then
      FileSpec = substr( FileSpec, 2, length( FileSpec) - 2)
   endif

   return FileSpec

; ---------------------------------------------------------------------------
defproc StripPath( Spec)
   return GetFileSpec( 'N', Spec)

; ---------------------------------------------------------------------------
defproc StripExt( Spec)
   return GetFileSpec( 'DPB', Spec)

; ---------------------------------------------------------------------------
defproc ChangeExt( FileSpec, NewExt)
   -- Strip optional leading dot in NewExt
   if leftstr( NewExt, 1) = '.' then
      NewExt = substr( NewExt, 2)
   endif
   return GetFileSpec( 'DPB', FileSpec)'.'NewExt

; ---------------------------------------------------------------------------
; This is e.g. for doubling quote chars in a cmd file line.
defproc DoubleCharInStr( Char, Str)
   pStart = 1
   do forever
      if Str == '' then
         leave
      endif
      p = verify( Str, Char, 'M', pStart)
      if not p then
         leave
      else
         Str = leftstr( Str, p - 1)''Char''substr( Str, p)
         pStart = p + 2
      endif
   enddo
   return Str

; ---------------------------------------------------------------------------
; Syntax: SaveAll
; Saves all files of the ring. Discards non-modified and temp files. Resets
; .modify for temp files. Brings unnamed files to the top.
defc SaveAll
   do once = 1 to 1
      getfileid startfid
      f = 0
      m = 0
      dprintf( 'RINGCMD', 'SaveAll')
      'VSyncCursor'
      do i = 1 to filesinring()  -- omit hidden files
         Filename = .filename
         fIsATempFile = (substr( Filename, 1, 1) = '.')
         fIsUnnamed   = (Filename = GetUnnamedFilename())
         if .modify then
            if fIsATempFile & not fIsUnnamed then
               .modify = 0
            else
               if fIsUnnamed then
                  refresh
               endif
               m = m + 1
               'Save'
               if rc = 0 then
                  f = f + 1
               endif
            endif
         endif
         nextfile
         getfileid fid
         if fid = startfid then
            leave
         endif
      enddo
      'PostMe ActivateFile' startfid
      if m > 0 then
         'SayError 'f' file(s) saved'
      else
         'SayError No file modified or just temp files in the ring'
      endif
   enddo

; ---------------------------------------------------------------------------
defc Quit, q
   do once = 1 to 1
      if not .visible then
         .modify = 0     -- so no "Are you sure?"
         'xcom quit'
         leave
      endif

      fTempFile = (leftstr( .filename, 1) = '.')
      fShell    = IsAShell()

compile if TRASH_TEMP_FILES
      if fTempFile then
         .modify = 0     -- so no "Are you sure?"
      endif
compile endif

      -- Workaround: This avoids unwanted window scrolling of the current file.
      'VSyncCursor'

      if .keyset = 'SPELL_KEYS' then  -- Dynamic spell-checking is on for this file;
         'dynaspell'                  -- toggle it off.
      endif

      -- Execute user macros
      'HookExecute quit'
      'HookExecuteOnce quitonce'

      if fShell then
         'ShellClose'
      else
         call QuitFile()
      endif
   enddo

; ---------------------------------------------------------------------------
; Used by defproc quit_file in E3EMUL.E and by defc shell_kill.
defc xcom_quit
   'xcom q'

; ---------------------------------------------------------------------------
; Save and quit.
defc File, f
   universal infolinerefresh

   infolinerefresh = 0
   do once = 1 to 1
      getfileid savefid
      's 'arg( 1)
      if rc then
         leave
      endif
      -- Don't close the wrong file - it may be already closed
      getfileid fid
      if fid <> savefid then
         leave
      endif
      .modify = 0
      'q'
   enddo
   infolinerefresh = 1

; ---------------------------------------------------------------------------
; Returns a found filename for a given Longname in Dir. It determines the
; filemask from Dir and a part of Longname and searches files until the
; Longname matches. Longname is handled case-insensitive.
; To avoid querying the longname for all files in a dir, the filemask
; includes the first 4 chars of the basename.
; Syntax:
;    FoundFilename = FindFilenameForLongname( [Filename [, Longname]])
defproc FindFilenameForLongname( Longname, Dir)
   SavedRc = rc
   LongBase = GetFileSpec( 'B', Longname)
   LongExt  = GetFileSpec( 'E', Longname)
   -- Avoid empty basename, e.g. for a temp. file like .NEPMD_INFO
   if LongBase = '' & LongExt <> '' then
      LongBase = '.'LongExt
      LongExt  = ''
   endif
   Drive = GetFileSpec( 'D', Dir)
   Path = Dir
   if rightstr( Path, 1) <> '\' then
      Path = Path'\'
   endif
   FileBase = MakeFilename( LongBase, Drive)
   FileBasePart = leftstr( FileBase, Min( 4, length( FileBase)))
   FileMask = Path''FileBasePart'*'

   -- Scan Dir for Longnames
   FoundFilename = ''
   UpLongname = upcase( Longname)
   Handle = ''
   NextFilename = ''
   do while NepmdGetNextFile( FileMask, Handle, NextFilename)
      NextLongname = NepmdQueryStringEa( NextFilename, '.LONGNAME')
      if NextLongname = '' then
         NextLongname = GetFileSpec( 'N', NextFilename)
      endif
      if upcase( NextLongname) = UpLongname then
         FoundFilename = NextFilename
         NepmdGetNextClose( Handle)
         leave
      endif
   enddo

   rc = SavedRc
   return FoundFilename

; ---------------------------------------------------------------------------
; If a match for a given Longname is found, it must exist.
; Syntax:
;    fExist = LongnameExist( [Filename [, Longname]])
defproc LongnameExist( Longname, Dir)
   fLongnameExists = (FindFilenameForLongname( Longname, Dir) <> '')
   return fLongnameExists

; ---------------------------------------------------------------------------
; Returns a unique longname for a passed filename and longname.
; Syntax:
;    UniqueLongname = MakeUniqueLongname( [Filename [, Longname]])
defproc MakeUniqueLongname
   Filename = arg( 1)
   if Filename = '' then
      Filename = .filename
   endif
   Dir = NepmdQueryFullname( Filename'\..')

   Longname = arg( 2)
   if Longname = '' then
      Longname = NepmdQueryStringEa( Filename, '.LONGNAME')
   endif
   if Longname = '' then
      Longname = GetFileSpec( 'N', Filename)
   endif

   -- Parse LongNum from LongPart
   nStart = 0
   LongPart = Longname
   lp = lastpos( ':', Longname)
   if lp then
      LongPart = leftstr( Longname, lp - 1)
      LongNum  = substr( Longname, lp + 1)
      --dprintf( 'Longname = "'Longname'" parsed into "'LongPart'" and "'LongNum'"')
      if IsUInt( LongNum) then
         nStart = LongNum
      endif
   endif

   if leftstr( LongPart, 1) = '.' & not pos( '.', substr( LongPart, 2)) then
      LongnameBase = LongPart
      LongnameExt  = ''
   else
      LongnameBase = GetFileSpec( 'B', LongPart)
      LongnameExt  = GetFileSpec( 'E', LongPart)
   endif
   UniqueLongname = LongPart

   -- Scan dir of Filename for Longnames
   n = nStart
   do forever
      if LongnameExist( UniqueLongname, Dir) then
         n = n + 1
         UniqueLongname = LongnameBase':'n
         if LongnameExt then
            UniqueLongname = UniqueLongname'.'LongnameExt
         endif
      else
         leave
      endif
   enddo

   -- Without path
   return UniqueLongname

; ---------------------------------------------------------------------------
; Checks if a Longname exists in a Dir. If yes, the file dialog is opened to
; enter another longname. Sets rc. Returns NewFilename. Loops until a
; non-existant longname is entered or on Cancel.
;    rc = -5   NewLongFilename = ''      Cancel or error
;    rc =  0   NewLongFilename = ''      Yes pressed or not found, overwrite
;    rc =  1   NewLongFilename = '...'   No pressed and NewLongFilename selected
; LongFilename means here: Longname with a prepended path, similar to a
; fully-qualified filename.
defproc SaveAsCheckLongnameExist( Longname, Dir)
   NewLongFilename = ''
   Path = Dir
   if rightstr( Path, 1) <> '\' then
      Path = Path'\'
   endif
   CheckLongFilename = Path''Longname
   fCheckAgain = 1
   rc = 0
   do forever

      if rc < 0 then
         --dprintf( '1: SaveAsCheckLongnameExist( 'Longname', 'Dir'): NewLongFilename = 'NewLongFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)
         leave
      endif
      if not fCheckAgain then
         --dprintf( '2: SaveAsCheckLongnameExist( 'Longname', 'Dir'): NewLongFilename = 'NewLongFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)
         leave
      endif
      CheckLongname = GetFileSpec( 'N', CheckLongFilename)
      -- NepmdQueryFullname works only for valid filenames
      CheckDir = GetFileSpec( 'DP', CheckLongFilename)
      if rightstr( CheckDir, 2) <> ':\' then
         CheckDir = leftstr( CheckDir, length( CheckDir) - 1)
      endif
      if not LongnameExist( CheckLongname, CheckDir) then
         --dprintf( '3: SaveAsCheckLongnameExist( 'Longname', 'Dir'): NewLongFilename = 'NewLongFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)
         leave
      endif
      fCheckAgain = 0

      Title = SAVE_AS__MSG
      Text = CheckLongFilename\n\n ||
             EXISTS_OVERLAY__MSG
      rcx = WinMessageBox( Title, Text,
                           MB_YESNOCANCEL + MB_DEFBUTTON2 + MB_WARNING + MB_MOVEABLE)
      if rcx = MBID_YES then
         rc = 0
      elseif rcx = MBID_NO then
         rc = 1
         -- Open file selector with NewName prefilled
         Title    = 'Longname exists - save with new name'
         FileMask = CheckLongFilename
         Flags    = 513
         NewLongFilename = FileDlg( Title, Filemask, Flags)
         if rc then                   -- Error
            rc = -5
         elseif not NewLongFilename then  -- Cancel
            rc = -5
         else                         -- OK
            rc = 1
            CheckLongFilename = NewLongFilename
            fCheckAgain = 1
         endif
      else
         rc = -5  -- sayerror( 'Access denied')
      endif
      --dprintf( '4: SaveAsCheckLongnameExist( 'Longname', 'Dir'): NewLongFilename = 'NewLongFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)

   enddo  -- forever
   --dprintf( '5: SaveAsCheckLongnameExist( 'Longname', 'Dir'): NewLongFilename = 'NewLongFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)

   -- Longname with prepended path
   return NewLongFilename

; ---------------------------------------------------------------------------
; Checks if a file exists. If yes, the file dialog is opened to enter another
; filename. Sets rc. Returns NewFilename. Loops until a non-existant filename
; is entered or on Cancel.
;    rc = -5   NewFilename = ''      Cancel or error
;    rc =  0   NewFilename = ''      Yes pressed or not found, overwrite
;    rc =  1   NewFilename = '...'   No pressed and NewFilename selected
defproc SaveAsCheckExist( Filename)
   NewFilename = ''
   CheckFilename = Filename
   fCheckAgain = 1
   rc = 0
   do forever

      if rc < 0 then
         --dprintf( '1: SaveAsCheckExist( 'Filename'): NewFilename = 'NewFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)
         leave
      endif
      if not fCheckAgain then
         --dprintf( '2: SaveAsCheckExist( 'Filename'): NewFilename = 'NewFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)
         leave
      endif
      if not Exist( CheckFilename) then
         --dprintf( '3: SaveAsCheckExist( 'Filename'): NewFilename = 'NewFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)
         leave
      endif
      fCheckAgain = 0

      Title = SAVE_AS__MSG
      Text = CheckFilename\n\n ||
             EXISTS_OVERLAY__MSG
      rcx = WinMessageBox( Title, Text,
                           MB_YESNOCANCEL + MB_DEFBUTTON2 + MB_WARNING + MB_MOVEABLE)
      if rcx = MBID_YES then
         rc = 0
      elseif rcx = MBID_NO then
         rc = 1
         -- Open file selector with NewName prefilled
         Title    = 'Filename exists - save with new name'
         FileMask = CheckFilename
         Flags    = 513
         NewFilename = FileDlg( Title, Filemask, Flags)
         if rc then                   -- Error
            rc = -5
         elseif not NewFilename then  -- Cancel
            rc = -5
         else                         -- OK
            rc = 1
            CheckFilename = NewFilename
            fCheckAgain = 1
         endif
      else
         rc = -5  -- sayerror( 'Access denied')
      endif
      --dprintf( '4: SaveAsCheckExist( 'Filename'): NewFilename = 'NewFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)

   enddo  -- forever
   --dprintf( '5: SaveAsCheckExist( 'Filename'): NewFilename = 'NewFilename', rc = 'rc', fCheckAgain = 'fCheckAgain)

   return NewFilename

; ---------------------------------------------------------------------------
; Syntax: SaveAs_Dlg [fAskIfExists]
;    fAskIfExists = 0: Omit the overlay dialog if file exists.
;    rc is set by defproc Save or, on error, by defproc SaveAs_Dlg.
/*
Ŀ
 what's it called: saveas_dlg      syntax:   saveas_dlg                     
                                                                            
 what does it do : ask EPM.EXE to pop up its "Save as" dialog box control.  
                   This is done by posting a EPM_POPOPENDLG message to the  
                   EPM Book window.                                         
                                                                            
                   (All EPM_EDIT_xxx messages are defined in the ETOOLKT    
                    PACKAGE available on PCTOOLS.)                          
                                                                            
 who and when    : Larry M.   6/12/91                                       

*/
defc SaveAs_Dlg
   do once = 1 to 1
      -- Optional arg, 0 => no EXIST_OVERLAY__MSG, used by def f2 if SMARTSAVE
      fAskIfExists = (arg( 1) <> 0)

      -- Select a filename for current file and change current title to it
      call SaveAs_Dlg( NewFilename, Type, fAskIfExists)
      --dprintf( 'SaveAs_Dlg: after call SaveAs_Dlg( 'NewFilename', 'Type', 'fAskIfExists')')
      if rc then  -- rc <> 0: Error
         leave
      endif

      -- Save current file
      fCalledBySaveAs = 1
      --dprintf( "SaveAs_Dlg: before call Save( '', "fCalledBySaveAs")")
      call Save( '', fCalledBySaveAs)

      AutosaveName = MakeTempName()
      if rc then
         call MoveFile( AutosaveName, MakeTempName())  -- Rename the autosave file
      else
         call DeleteFile( AutosaveName)
      endif
   enddo

; ---------------------------------------------------------------------------
; Calls the FileDlg via ERESSaveas of ERES603.DLL to select a filename for
; current file.
; Both args Filename and Type are passed by call-by-reference.
;    rc =  0  Success
;    rc <> 0  Error
defproc SaveAs_Dlg( var NewLongFilename, var Type)
   rc = 1
   do once = 1 to 1
      StartFilename = .filename
      StartLongname = Get_EAT_ASCII_Value( '.LONGNAME')
      if StartLongname = '' then
         StartLongname = GetFileSpec( 'N', StartFilename)
      endif

      -- Optional 3rd arg, 0: no EXIST_OVERLAY__MSG, used by def f2 if SMARTSAVE
      fAskIfExists = (arg( 3) <> 0)

      fUsePrevDir = 0
      fTempFile = (leftstr( .filename, 1) = '.')
      if fTempFile then
         KeyPath = '\NEPMD\User\StartDir\SaveasDlg\Type'
         opt = QueryConfigKey( KeyPath)
         if opt = 0 then
            fUsePrevDir = 1
         endif
      endif

      if fUsePrevDir then
         LastSavedFile = GetHistoryEntry( 'Save', 1)
         FilePath = GetFileSpec( 'DP', LastSavedFile)
      else
         FilePath = GetFileSpec( 'DP', .filename)
      endif

      do once2 = 1 to 1
         NewLongFilename = .filename

         if FilePath = '' then
            leave
         endif
         if leftstr( NewLongFilename, 1) = '.' then  -- if a temp file
            leave
         endif
         if StartLongname = '' then
            leave
         endif

         NewLongFilename = FilePath''StartLongname
      enddo
      --dprintf( 'SaveAs_Dlg (proc) 1: NewLongFilename = 'NewLongFilename)

      -- Open FileDlg via ERESSaveas. Start with current filename. On error,
      -- open it again with empty name. For a FAT source file, its .LONGNAME
      -- is used in the dialog.
      NewLongFilename = leftstr( NewLongFilename, 255, \0)
      Type = copies( \0, 255)
      do i = 1 to 2
         rc = dynalink32( ERES2_DLL,                -- library name
                          'ERESSaveas',             -- function name
                          GethWndC( EPMINFO_EDITFRAME)  ||
                          GethWndC( APP_HANDLE)         ||
                          address( NewLongFilename)            ||
                          address( Type))
         -- Return codes:  0 = OK; 1 = memory problem; 2 = bad string; 3 = couldn't load control from DLL
         if rc = 2 then      -- File dialog didn't like the .filename;
            -- Try again with no file name
            NewLongFilename = copies( \0, 255)
            iterate
         endif
         leave
      enddo

      parse value NewLongFilename with NewLongFilename \0
      parse value Type with Type \0
      -- NewLongFilename returned from ERESSaveas is always enquoted
      NewLongFilename = StripQuotesFileSpec( NewLongFilename)
      --dprintf( 'SaveAs_Dlg (proc) 2: NewLongFilename = 'NewLongFilename', rc = 'rc)

      if NewLongFilename = '' then
         rc = -275  -- sayerror( 'Missing filename')
         leave
      elseif rc then
         leave
      endif

      if upcase( NewLongFilename) = upcase( FilePath''StartLongname) then
         crc = 0
      else
         crc = 1
      endif
      -- If filename, entered with the first FileDlg, exists, allow for to
      -- select another filename
      if fAskIfExists then
         NewLongname = GetFileSpec( 'N', NewLongFilename)
         -- NepmdQueryFullname works only for valid filenames
         --NewDir      = NepmdQueryFullName( NewLongFilename'\..')
         NewDir = GetFileSpec( 'DP', NewLongFilename)
         if rightstr( NewDir, 2) <> ':\' then
            NewDir = leftstr( NewDir, length( NewDir) - 1)
         endif
         --dprintf( 'SaveAs_Dlg (proc) 3: NewDir = 'NewDir)
         ChangedLongFilename = SaveAsCheckLongnameExist( NewLongname, NewDir)
         crc = rc
         --dprintf( 'SaveAs_Dlg (proc) 4: 'ChangedLongFilename' = SaveAsCheckLongnameExist( 'NewLongname', 'NewDir'), rc = 'rc)
         if crc < 0 then      -- Cancel or error
            leave
         elseif crc = 0 then  -- Overwrite
            -- nop
            if ChangedLongFilename then
               NewLongFilename = ChangedLongFilename
            endif
         elseif crc = 1 then  -- Other name selected
            --dprintf( 'SaveAs_Dlg: rc = 1, change current filename '.filename' -> 'NewLongFilename)
            -- Change returned NewLongFilename
            NewLongFilename = ChangedLongFilename
         endif
      endif

      -- Build valid filename from entered NewLongFilename
      NewDrive = GetFileSpec( 'D', NewLongFilename)
      NewPath  = GetFileSpec( 'DP', NewLongFilename)
      NewLongname  = GetFileSpec( 'N', NewLongFilename)
      NewShortname = MakeFilename( NewLongname, NewDrive)
      NewShortFilename = NewPath''NewShortname
      --dprintf( 'SaveAs_Dlg (proc) 5: NewShortname = 'NewShortname', NewLongname = 'NewLongname', StartFilename = 'StartFilename)

      if crc = 1 then
         --dprintf( 'SaveAs_Dlg (proc) 6: Change .filename: Start with NewShortFilename = 'NewShortFilename)
         .filename = MakeUniqueFilename( NewShortFilename)
         --dprintf( 'SaveAs_Dlg (proc) 7: Changed .filename to unique filename = '.filename)
      elseif crc = 0 /*& ChangedLongFilename*/ then
         NewDir = NewPath
         if rightstr( NewDir, 2) <> ':\' then
            NewDir = leftstr( NewDir, length( NewDir) - 1)
         endif
         FoundFilename = FindFilenameForLongname( NewLongname, NewDir)
         if FoundFilename then
            .filename = FoundFilename
            --dprintf( 'SaveAs_Dlg (proc) 8: Changed .filename to FoundFilename = '.filename)
         else
            .filename = NewShortFilename
            --dprintf( 'SaveAs_Dlg (proc) 9: Changed .filename to NewShortFilename = '.filename)
         endif
      endif

      -- Must delete EA from .eaarea first to refresh the displayed longname
      call Delete_Ea( '.LONGNAME')
      if NewLongname <> NewShortname & NewLongname then
         --dprintf( '* SaveAs_Dlg (proc) 10: Add .LONGNAME = 'NewLongname)
         'Add_Ea .LONGNAME' NewLongname
      endif

      -- Refresh .TYPE on success
      if Type then
          call Delete_Ea( '.TYPE')
         'Add_Ea .TYPE' Type
      endif

      rc = 0
   enddo  -- once

   return

; ---------------------------------------------------------------------------
/*
Ŀ
 what's it called: opendlg         syntax:   opendlg [EDIT|GET]             
                                                                            
 what does it do : ask EPM.EXE to pop up its internal message box control.  
                   This is done by posting a EPM_POPOPENDLG message to the  
                   EPM Book window.                                         
                   If a file is selected, by default, it will be presented  
                   in a new window.  If the 'EDIT' option is specified the  
                   file specified will be opened in the active edit window. 
                                                                            
                   (All EPM_EDIT_xxx messages are defined in the ETOOLKT    
                    PACKAGE available on PCTOOLS.)                          
                                                                            
 who and when    : Jerry C.   2/27/89                                       

*/
defc OpenDlg
   universal ring_enabled
   universal app_hini

   if upcase( arg( 1)) = 'EDIT' then
      style = 1  -- EDIT
   elseif upcase( arg( 1)) = 'GET' then
      style = 2  -- GET
   else
      style = 0  -- OPEN
   endif

   ThisFilename = .filename
   ThisDir = ''
   lp = lastpos( '\', ThisFilename)
   if lp > 1 then
      ThisDir = leftstr( ThisFilename, lp - 1)
   endif

   PrevFileSelected = queryprofile( app_hini, 'ERESDLGS', 'LASTFILESELECTED')
   PrevDir = ''
   lp = lastpos( '\', PrevFileSelected)
   if lp > 1 then
      PrevDir = leftstr( PrevFileSelected, lp - 1)
   endif

   WorkDir = directory()

   KeyPath = '\NEPMD\User\StartDir\OpenDlg\Type'
   opt = QueryConfigKey( KeyPath)
   -- 0 = previous dir
   -- 1 = work dir
   -- 2 = dir of current file

   if opt = 2 & pos( ':\', ThisFilename) = 0 then
      opt = 1
   endif

   -- Make the standard file dialog select the configured dir.
   -- The ini key LASTFILESELECTED is read internally by the file dialog adapted
   -- for EPM, but not by EPM's 'Open' or 'Add file' dialog.
   fWriteLastFileSelected = 0
   LastFileSelected = PrevFileSelected
   if opt = 0 then      -- 0 = previous dir
      -- nop
   elseif opt = 1 then  -- 1 = work dir
      LastFileSelected = ''
      fWriteLastFileSelected = 1
   elseif opt = 2 then  -- 2 = dir of current file
      LastFileSelected = ThisFilename
      fWriteLastFileSelected = 1
   endif

   if fWriteLastFileSelected then
      call setprofile( app_hini, 'ERESDLGS', 'LASTFILESELECTED', LastFileSelected)
   endif

   -- Make the EPM 'Open' or 'Add file' dialog select the configured dir.
   -- Different from the file dialog, the dir of the 'Add file' dialog can only
   -- be configured by temporarily changing the work dir. The default
   -- behavior is to open the dir configured by the program object.
   -- Note: The startup option /o suppresses this dialog and opens the
   --       file dialog instead.
   fChangeDir = 0
   NewDir = ''
   if opt = 0 then      -- 0 = previous dir
      NewDir = PrevDir
      fChangeDir = 1
   elseif opt = 1 then  -- 1 = work dir
      -- nop
   elseif opt = 2 then  -- 2 = dir of current file
      NewDir = ThisDir
      fChangeDir = 1
   endif

   if fChangeDir then
      NextDir = Directory2( NewDir)
   endif

   call windowmessage( 0, getpminfo( APP_HANDLE),
                       5126,               -- EPM_POPOPENDLG
                       ring_enabled,
                       style * 65536)  -- OPEN = 0; EDIT = 1; GET = 2

   -- The change back to the previous dir must be posted to work reliable.
   if fChangeDir then
      'PostMe Cd' WorkDir
   endif

; ---------------------------------------------------------------------------
; Syntax: Filename = FileDlg( title[, filemask[, flags[, OKbuttontext]]])
defproc FileDlg( Title)

   do once = 1 to 1
      -- Return value
      SelectedFilename = ''

      Filemask = arg( 2)
      if Filemask = '' then
         Filemask = '*'
      endif

      Flags = arg( 3)
      if Flags = '' then
         Flags = 257  -- FDS_CENTER + FDS_OPEN_DIALOG
      endif
      /*  -- pmstddlg.h
      #define FDS_CENTER            0x00000001L /*    1 Center within owner wnd   */
      #define FDS_FILTERUNION       0x00000004L /*    4 Use union of filters      */
      #define FDS_OPEN_DIALOG       0x00000100L /*  256 Select Open dialog        */
      #define FDS_SAVEAS_DIALOG     0x00000200L /*  512 Select SaveAs dialog      */
      */

      OkButtonText = arg( 4)

      Size         = 328    -- Size of FILEDLG struct
      Title        = Title\0
      FileDlg = atol( Size) || atol( Flags) || copies( \0, 12) ||
                address( Title) || copies( \0, Size - 24)
      FileDlg = overlay( Filemask, FileDlg, 53)          -- Insert Filemask
      if OkButtonText then
         OkButtonText = OkButtonText\0
         pOkButtonText = address( OkButtonText)
         FileDlg = overlay( pOkButtonText, FileDlg, 25)  -- Insert OkButtonText
      endif

      hWndDlg = dynalink32( 'PMCTLS',
                            'WINFILEDLG',
                            atol( 1)                      ||  -- Parent, file dialog is clipped by parent size
                            GethWndC( EPMINFO_EDITFRAME)  ||  -- Owner
                            address( FileDlg))
      rc = ltoa( substr( FileDlg, 17, 4), 10)
      if not hWndDlg then
         leave
      endif

      parse value substr( FileDlg, 53) with Filename \0
      Button = ltoa( substr( FileDlg, 13, 4), 10)

      if Button = 1 & Filename <> '' then
         SelectedFilename = Filename
      endif

   enddo
   return SelectedFilename

; ---------------------------------------------------------------------------
; Syntax: FileDlg title[, cmd[, filemask[, flags]]]
; Opens a standard WinFileDlg and executes cmd with the selected filename.
; If cmd contains none of the WPS parameter macros, filename is appended
; to cmd with a space in between. Otherwise the WPS parameter macro is
; resolved with filename.
; For interaction with EPM REXX commands, the cmd can be specified as
; "SetUserstring" (not case-sensitive). Then the selected filename is set to
; the .userstring field value. It can be extracted with 'extract /userstring'
; and then processed further with userstring.1 by an EPM REXX command.
; Used by:
;    EDIT.E: defc OpenBinDlg
; When being used within an Open cmd, FileDlg sometimes makes the following
; messages crash EPM. An entry in POPUPLOG.OS2 is created then.
; After a thread in comp.os.os2.programmer.tools, 1997-02-11.
; | Larry Margolis: An example of using a file dialog from the macros.
; | Modified from code by Bernhard Bablok.
; Changed.
defc FileDlg
   MacroList = '%**P %**D %**F %**N %**E %*'

   parse arg Title ',' Cmd ',' Rest
   Rest = strip( Rest)
   if leftstr( Rest, 1) = '"' then
      parse value Rest with '"'Filemask'"' Flags Rest2
   else
      parse value Rest with Filemask Flags Rest2
   endif
   if leftstr( Rest2, 1) = '"' then
      parse value Rest2 with '"'OkButtonText'"' .
   else
      parse value Rest2 with OkButtonText .
   endif
   Title        = strip( Title)
   Cmd          = strip( Cmd)
   Filemask     = strip( Filemask)
   Flags        = strip( Flags)
   OkButtonText = strip( OkButtonText)

   if Cmd = '' then
      Cmd = 'e'
   endif

   Filename = FileDlg( Title, Filemask, Flags, OkButtonText)

   if rc then
      'SayError Error: WinFileDlg returned rc = 'rc'.'
   elseif Filename = '' then
      if upcase( Cmd) = 'SETUSERSTRING' then
         -- For interaction with .erx files: set selected filename to .userstring
         .userstring = ''
      endif
   elseif Filename <> '' then
      if upcase( Cmd) = 'SETUSERSTRING' then
         -- For interaction with .erx files: set selected filename to .userstring
         .userstring = Filename
      else
         -- Check if a %* macro is specified
         fWpsParamMacro = 0
         do w = 1 to words( MacroList)
            Macro = word( MacroList, w)
            if pos( Macro, Cmd) then
               fWpsParamMacro = 1
               leave
            endif
         enddo
         if fWpsParamMacro then
            -- Execute it
            ExpandWpsParameterMacros( Cmd, Filename)
         else
            -- Execute Cmd with Filename appended
            Cmd Filename
         endif
      endif
   endif

; ---------------------------------------------------------------------------
; Locks file only if modified, unlocks if .modify = 0
; Todo: disable this if it failed once and for slow drives.
defmodify
   if .visible & leftstr( .filename, 1) <> '.' then
      KeyPath = '\NEPMD\User\Lock\OnModify'
      LockOnModify = QueryConfigKey( KeyPath)
      if LockOnModify = 1 then
         if .modify > 0 & .lockhandle = 0 & .readonly = 0 then
            'Lock'
         elseif .modify = 0 & .lockhandle > 0 then
            'Unlock'
         endif
      endif
   endif
   --dprintf( '.modify = '.modify', .lockhandle = '.lockhandle', .readonly = '.readonly)

; ---------------------------------------------------------------------------
; Can be used to toggle the lock state.
defc lock_on_modify
   if wordpos( upcase( arg( 1)), '0 OFF') then
      on = 0
   else
      on = 1
   endif
   if .visible & leftstr( .filename, 1) <> '.' then
      if on then
         if .modify > 0 & .lockhandle = 0 & .readonly = 0 then
            'Lock'
         endif
      else
         if .lockhandle <> 0 & .readonly = 0 then
            'Unlock'
         endif
      endif
   endif

; ---------------------------------------------------------------------------
defc Lock
   if arg( 1) <> '' then
      'e 'arg( 1)
      if rc & rc <> -282 then  --sayerror( 'New file')
         return 1
      endif
   endif
   if .readonly then
      -- We don't need a msg here
      -- sayerror '"'.filename'" is readonly -- not locked'
   else
      if Exist( .filename) then  -- added check for exist to disable 0 byte files created by lock
         rc = Lock( 'W')  -- deny Write
      endif
   endif

; ---------------------------------------------------------------------------
; Syntax: call lock( [<deny_mode>])
;         <deny_mode> = 'W' or 'RW', default is 'RW'
/*
   -- from bsedos.h:
   /* DosOpen/DosSetFHandState mode flags */
   #define OPEN_ACCESS_READONLY           0x0000  /* ---- ---- ---- -000 */
   #define OPEN_ACCESS_WRITEONLY          0x0001  /* ---- ---- ---- -001 */
   #define OPEN_ACCESS_READWRITE          0x0002  /* ---- ---- ---- -010 */
   #define OPEN_SHARE_DENYREADWRITE       0x0010  /* ---- ---- -001 ---- */
   #define OPEN_SHARE_DENYWRITE           0x0020  /* ---- ---- -010 ---- */
   #define OPEN_SHARE_DENYREAD            0x0030  /* ---- ---- -011 ---- */
   #define OPEN_SHARE_DENYNONE            0x0040  /* ---- ---- -100 ---- */
   #define OPEN_FLAGS_NOINHERIT           0x0080  /* ---- ---- 1--- ---- */
   #define OPEN_FLAGS_NO_LOCALITY         0x0000  /* ---- -000 ---- ---- */
   #define OPEN_FLAGS_SEQUENTIAL          0x0100  /* ---- -001 ---- ---- */
   #define OPEN_FLAGS_RANDOM              0x0200  /* ---- -010 ---- ---- */
   #define OPEN_FLAGS_RANDOMSEQUENTIAL    0x0300  /* ---- -011 ---- ---- */
   #define OPEN_FLAGS_NO_CACHE            0x1000  /* ---1 ---- ---- ---- */
   #define OPEN_FLAGS_FAIL_ON_ERROR       0x2000  /* --1- ---- ---- ---- */
   #define OPEN_FLAGS_WRITE_THROUGH       0x4000  /* -1-- ---- ---- ---- */
   #define OPEN_FLAGS_DASD                0x8000  /* 1--- ---- ---- ---- */
   #define OPEN_FLAGS_NONSPOOLED          0x00040000
   #define OPEN_SHARE_DENYLEGACY       0x10000000   /* 2GB */
   #define OPEN_FLAGS_PROTECTED_HANDLE 0x40000000
*/
defproc Lock
   DenyReadWrite = 146 -- $92 = $80 + $10 + $2 = OPEN_FLAGS_NOINHERIT + OPEN_SHARE_DENYREADWRITE + OPEN_ACCESS_READWRITE
;  DenyWrite     = 160 -- $A0 = $80 + $20 + $0 = OPEN_FLAGS_NOINHERIT + OPEN_SHARE_DENYWRITE + OPEN_ACCESS_READONLY   <-- doesn't work
;  DenyWrite     = 161 -- $A1 = $80 + $20 + $1 = OPEN_FLAGS_NOINHERIT + OPEN_SHARE_DENYWRITE + OPEN_ACCESS_WRITEONLY  <-- works
   DenyWrite     = 162 -- $A1 = $80 + $20 + $2 = OPEN_FLAGS_NOINHERIT + OPEN_SHARE_DENYWRITE + OPEN_ACCESS_READWRITE  <-- works, use this
   DenyMode = upcase( arg( 1))
   If DenyMode = '' then
      Deny = DenyReadWrite  -- like in standard EPM
   elseif DenyMode = 'W' then
      Deny = DenyWrite
   elseif DenyMode = 'RW' then
      Deny = DenyReadWrite
   else
      --sayerror 'defproc lock: Unknown arg "'DenyMode'"'
      --return
      Deny = arg( 1)
   endif
   file = .filename\0
   newhandle = '????'
   actiontaken = atol( 1)   -- file exists
   result = dynalink32( 'DOSCALLS',
                        '#273',             -- Dos32Open
                        address( file)         ||
                        address( newhandle)    ||
                        address( actiontaken)  ||
                        atol( 0)    ||       -- file size
                        atol( 0)    ||       -- file attribute
                        atol( 17)   ||       -- open flag; open if exists, else create
                        atol( Deny) ||       -- openmode; deny Read/Write or Write
                        atol( 0), 2)
   if result then
      if result = 32 then
         'SayError 'ACCESS_DENIED__MSG '-' MAYBE_LOCKED__MSG' rc = ERROR_SHARING_VIOLATION.'
         return result
      elseif result = 206 then  -- 206 = ERROR_FILENAME_EXCED_RANGE
         return result
      else
         --messageNwait( 'DOSOPEN' ERROR__MSG result NOT_LOCKED__MSG)
         return result
      endif
   endif
   if newhandle  = \0\0\0\0 then  -- Handle of 0 - bad news
      newhandle2 = \255\255\255\255
      result = dynalink32( 'DOSCALLS',
                           '#260',      -- Dos32DupHandle
                           newhandle ||
                           address( newhandle2), 2)
      call dynalink32( 'DOSCALLS',      -- free the original handle
                       '#257',          -- Dos32Close
                       newhandle, 2)
      if result then
         MessageNWait( 'DosDupHandle' ERROR__MSG result NOT_LOCKED__MSG)
         return result
      endif
      newhandle = newhandle2
   endif
   .lockhandle=ltoa( newhandle, 10)
   return 0

; ---------------------------------------------------------------------------
defc Unlock
   parse arg file
   if file = '' then
      getfileid fileid
   else
      getfileid fileid, file
      if fileid == '' then
         'SayError "'file'"' DOES_NOT_EXIST__MSG
         return 1
      endif
   endif
   call Unlock( fileid)

; ---------------------------------------------------------------------------
defproc Unlock( fileid)
   if fileid.lockhandle = 0 then
      'SayError 'fileid.filename NOT_LOCKED__MSG
      return 1
   endif
   result = dynalink32( 'DOSCALLS',    -- Free the original handle
                        '#257',                    -- Dos32Close
                        atol( fileid.lockhandle), 2)
   if result then
      'SayError DosClose 'ERROR_NUMBER__MSG result
   else
      fileid.lockhandle = 0
   endif
   return result

; ---------------------------------------------------------------------------
; This doesn't change rc.
defproc LoadHiddenFile( Name)
   fid = ''
   savedrc = rc
   getfileid curfid

   do once = 1 to 1
      'xcom e /c 'EnquoteFileSpec( Name)
      if rc <> -282 then  -- -282 = sayerror( "New file")
         'SayError 'ERROR__MSG rc BAD_TMP_FILE__MSG sayerrortext( rc)
         leave
      endif
      fBrowse = browse()     -- query current state
      if fBrowse then
         call browse( 0)
      endif
      getfileid fid
      .autosave = 0
      .visible = 0
      pActivateFile( curfid)
   enddo

   rc = savedrc
   return fid

; ---------------------------------------------------------------------------
; This 'xcom quit' has to be posted when called during file loading.
; Otherwise wrong messages occur:
; -  Invalid fileid
; -  The JustInstalled flag can't be read in MAIN.E from NEPMD.INI.
; This doesn't change rc.
defproc CloseHiddenFile( fid)
   getfileid curfid
   startfid = arg( 2)
   if startfid = '' then
      startfid = curfid
   endif
   savedrc = rc

   do once = 1 to 1
      if fid = '' then
         leave
      endif
      if not validatefileid( fid) then
         leave
      endif

      pActivateFile( fid)
      .modify = 0
      'xcom quit'
      -- Quit may switch to another file after execution, so always restore
      -- startfid.
      -- Must delay this until ProcessFilesLoaded is done, if a file was
      -- loaded.
      'AfterStartup ActivateFile 'startfid
   enddo

   rc = savedrc
   return

defc CloseHiddenFile
   parse arg fid startfid .
   call CloseHiddenFile( fid, startfid)

; ---------------------------------------------------------------------------
; This doesn't change rc.
defproc SaveCloseHiddenFile( fid, Filename)
   startfid = arg( 3)
   savedrc = rc

   do once = 1 to 1
      if fid = '' then
         leave
      endif
      if not validatefileid( fid) then
         leave
      endif
      pActivateFile( fid)
      .modify = 0
      if Exist( Filename) then
         DeleteFile( Filename)
      endif
      'xcom save' Filename
      'xcom quit'

      if startfid = '' then
         leave
      endif
      -- Must delay this until ProcessFilesLoaded is done, if a file was
      -- loaded.
      'AfterStartup ActivateFile 'startfid
   enddo

   rc = savedrc
   return

; ---------------------------------------------------------------------------
defproc AppendHiddenFile( fid, LineStr)
   fUnique = (arg( 3) = 1)

   fWrite = 1
   if fUnique then
      -- Check if LineStr was already added
      fFound = LineFoundInHiddenFile( fid, LineStr)
      if fFound then
         fWrite = 0
      endif
   endif

   if fWrite = 1 then
      -- Append LineStr
      insertline LineStr, fid.last + 1, fid
   endif

   return

; ---------------------------------------------------------------------------
; Checks if submitted LineStr is already present in fid. Not case-sensite
; to handle filespecs.
defproc LineFoundInHiddenFile( fid, LineStr)
   fFound = 0

   -- Check if LineStr was already added
   do l = 1 to fid.last
      getline AddedName, l, fid
      if upcase( LineStr) = upcase( AddedName) then
         fFound = 1
         leave
      endif
   enddo

   return fFound

; ---------------------------------------------------------------------------
; This can be used for debugging. Either a fileid or a filename can be passed
; to it.
defproc ShowHiddenFile( fid)
   if not IsNum( fid) then
      Filename = fid
      getfileid fid, Filename
   endif
   if validatefileid( fid) then
      activatefile fid
      .visible = 1
      .modify = 0
      'ProcessLoad'
   endif
   return

defc ShowHiddenFile
   fid = strip( arg( 1))
   call ShowHiddenFile( fid)

; ---------------------------------------------------------------------------
; Extended to allow for specifying a filename.
; Syntax: pActivateFile( <fid>)
;         pActivateFile( <filename>)
;         pActivateFile( <fullfilename>)
; Sets rc to the rc of activatefile. Avoids error message of it.
defproc pActivateFile( fid)
   if not IsNum( fid) then
      Filename = fid
      getfileid fid, Filename
   endif

   if fid <> '' then
      -- Check if file to be activated is still in ring and visible
      if wordpos( validatefileid( fid), '1 2') then
         activatefile fid
      else
         rc = -260  -- Invalid fileid
      endif
   endif
   return

defc ActivateFile
   call pActivateFile( strip( arg( 1)))

; ---------------------------------------------------------------------------
compile if not defined( EPM_EDIT_LOGAPPEND)
const
   EPM_EDIT_LOGAPPEND = 5496
compile endif

defproc WriteFile( OutFile, Text)
   if Exist( OutFile) then
      call DeleteFile( OutFile)
   endif
   call AppendFile( OutFile, Text)
   return

defproc AppendFile( OutFile, Text)
   TextZ = Text\0
   OutFileZ = OutFile\0
   call windowmessage( 1, getpminfo( EPMINFO_EDITFRAME),
                       EPM_EDIT_LOGAPPEND,
                       ltoa( offset( OutFileZ)''selector( OutFileZ), 10),
                       ltoa( offset( TextZ)''selector( TextZ), 10))
   return

; ---------------------------------------------------------------------------
; Creates a temp. CmdFile with a Cmd and executes it. Sets rc to rc of the
; executed Cmd.
defproc RunCmdFile( CmdFile, Cmd)
   call WriteFile( CmdFile, Cmd)

   EnqCmdFile = EnquoteFileSpec( CmdFile)

   -- 'cmd' is required here. Otherwise rc = -274 (Unknown command) is returned.
   quietshell 'cmd /c 'EnqCmdFile
   CmdRc = rc

   call DeleteFile( CmdFile)

   rc = CmdRc
   return

; ---------------------------------------------------------------------------
defc Readonly
   uparg = upcase( arg( 1))
   if (uparg = ON__MSG or uparg = 1) then
      .readonly = 1
   elseif (uparg = OFF__MSG or uparg = '0') then
      .readonly = 0
   elseif (uparg = '' or uparg = '?') then
      'SayError 'READONLY_IS__MSG word( OFF__MSG ON__MSG, .readonly + 1)
   else
      'SayError 'INVALID_ARG__MSG ON_OFF__MSG'/?)'
      stop
   endif

; ---------------------------------------------------------------------------
defc ReadonlyAttrib
   uparg = upcase( arg( 1))
   if (uparg = ON__MSG or uparg = 1) then
      quietshell 'attrib +r 'EnquoteFileSpec( .filename)
   elseif (uparg = OFF__MSG or uparg = '0') then
      quietshell 'attrib -r 'EnquoteFileSpec( .filename)
   elseif (uparg = '' or uparg = '?') then
      'SayError Read-only file attribute is:' word( OFF__MSG ON__MSG, GetReadOnly() + 1)
   else
      'SayError 'INVALID_ARG__MSG ON_OFF__MSG'/?)'
      stop
   endif

; ---------------------------------------------------------------------------
defc EnableReadonly
   if wordpos( upcase( arg( 1)), '0 OFF') then
      on = 0
   else
      on = 1
   endif
   if on then
      .readonly = GetReadonly()
   else
      .readonly = 0
   endif

; ---------------------------------------------------------------------------
; Determines .readonly field from file attribute, if enabled.
defc ReadonlyFromAttrib
   KeyPath = '\NEPMD\User\Readonly'
   RespectReadonly = QueryConfigKey( KeyPath)
   if RespectReadonly then
      -- Get file attributes to set the .readonly field var
      -- If .readonly is set and = 1, then the %M template pattern for the
      -- status bar will show 'Read-only' and disables any modification,
      -- until 'readonly 0' or 'readonly off' is set.
      .readonly = GetReadonly()
   else
      .readonly = 0
   endif

; ---------------------------------------------------------------------------
defproc GetReadonly
   readonly = ''
   savedrc = rc
/*
   -- 1) using NepmdQueryPathInfo
   attr = NepmdQueryPathInfo( .filename, 'ATTR')
   if rc then  -- file doesn't exist
      --'SayError Attributes for "'Filename'" can''t be obtained, rc = 'rc
   elseif length( attr) = 5 then
      readonly = (substr( attr, 5, 1) = 'R')
   endif
*/
   -- 2) using QueryFileMode
   Attrib = QueryFileMode( .filename)  -- DosQFileMode
   if not rc then   -- file exists
      readonly = (Attrib // 2)
   endif

   rc = savedrc
   return readonly

; ---------------------------------------------------------------------------
; Outdated. Used before ReadOnly existed.
;
; Syntax: Browse off/on/0/1
;
; specifies whether E should allow text to be altered (normal editing mode)
; or whether all text is read-only.
;
; Issuing Browse with '?' or no argument returns the current setting.
;
; The function browse() takes an optional argument 0/1.  It always returns
; the current setting.  So you can query the current setting without changing
; it by giving no argument.
defc Browse
   uparg = upcase( arg( 1))
   if (uparg = ON__MSG or uparg = 1) then
      curbrowse = browse( 1)
   elseif (uparg = OFF__MSG or uparg = '0') then
      curbrowse = browse( 0)
   elseif (uparg = '' or uparg = '?') then
      curbrowse = browse()     -- query current state
      'SayError 'BROWSE_IS__MSG word( OFF__MSG ON__MSG, curbrowse + 1)
   else
      'SayError 'INVALID_ARG__MSG ON_OFF__MSG'/?)'
   endif

; ---------------------------------------------------------------------------
; Changes drive. Releases previous directory first (changes to root dir).
; Accepts drives and dirs as arg.
defproc ChDrive
   NewDir = arg( 1)
   ChangedDir = ''

   do once = 1 to 1
      if NewDir = '' then
         leave
      endif
      if substr( NewDir, 2, 1) <> ':' then
         leave
      endif
      CurDrive = upcase( substr( directory(), 1, 2))
      NewDrive = upcase( substr( NewDir, 1, 2))
      if NewDrive = CurDrive then
         leave
      endif
      -- Change to root
      call directory( '\')
      -- Change drive
      ChangedDir = directory( NewDrive)
   enddo

   return ChangedDir

defc ChDrive
   ChangedDir = ChDrive( StripQuotesFileSpec( arg( 1)))

; ---------------------------------------------------------------------------
; Changes drive and directory. Releases previous directory first (changes to
; root dir).
; This should be used e.g. to change the directory on every defselect.
defproc Directory2
   NewDir = arg( 1)

   call ChDrive( NewDir)

   -- Change dir and return full name on success
   ChangedDir = directory( NewDir)

   -- Save last changed dir
   KeyPath = '\NEPMD\User\ChangeWorkDir\Last'
   call WriteConfigKey( KeyPath, NewDir)

   return ChangedDir

; ---------------------------------------------------------------------------
; Changes drive and directory. Releases previous directory first (changes to
; root dir). This should be used e.g. to change the directory on every
; defselect. Compared to Directory2, the 'cd' command is used to display
; messages and to remember the new directory for restoring path, if
; activated. On a successful directory change and if the current directory
; is specified, no message is given.
defc cd, cdd, WorkDir
   do once = 1 to 1
      rc = 0
      arg1 = arg( 1)
      arg1 = StripQuotesFileSpec( arg1)
      if arg1 <> '' then
         NewDir = Directory2( arg1)  -- Returned value is after dir change
         if NewDir = '' then
            'SayError Error: Directory not changed.'
            leave
         endif
      else
         'SayError 'CUR_DIR_IS__MSG directory()
      endif
   enddo

; ---------------------------------------------------------------------------
; Changes drive and directory. If a filename was specified, the directory of
; that file is used instead.
defc xcd
   NewDir = arg( 1)
   NewDir = StripQuotesFileSpec( NewDir)

   -- Handle dir and file specs. For a file spec, use its dir to change to.
   do once = 1 to 1
      if NewDir = '' then
         leave
      endif
      if NepmdDirExists( NewDir) then
         leave
      endif
      if NepmdFileExists( NewDir) then
         -- Get 1 level back
         NewDir = leftstr( NewDir, lastpos( '\', NewDir) - 1)
      endif
   enddo

   -- Change drive and dir. Release previous directory first.
   'cd' NewDir

; ---------------------------------------------------------------------------
; Changes drive and directory. Releases previous directory first (changes to
; root dir). Gives a message.
defc CDDlg
   CurDir = directory()
   Title = 'Enter new work directory'
   Text  = CUR_DIR_IS__MSG CurDir
   Text  = Text''copies( ' ', Max( 100 - length( Text), 0))
   Entry = ''
   parse value EntryBox( Title,
                         '',
                         Entry,
                         0,
                         240,
                         atoi( 1) || atoi( 0) || atol( 0) ||
                         Text) with button 2 NewDir \0
   NewDir = strip( NewDir)
   if button = \1 & NewDir <> '' & NewDir <> CurDir then
      'xcd' NewDir  -- Accept a dir or a file
      if rc then
         'SayError Directory not changed. rc = 'rc
      else
         'cd'       -- Show current dir
      endif
   endif

; ---------------------------------------------------------------------------
; Helper to set startup dir for EPM program objects.
defc StartupDirDlg
   KeyPath = '\NEPMD\User\ProgramObjects\StartupDir'
   CurStartupDir = QueryConfigKey( KeyPath)
   Title = 'Enter new startup directory for WPS objects'
   Text  = 'Current startup directory is 'CurStartupDir
   Text  = Text''copies( ' ', Max( 100 - length( Text), 0))
   Entry = ''
   parse value EntryBox( Title,
                         '',
                         Entry,
                         0,
                         240,
                         atoi( 1) || atoi( 0) || atol( 0) ||
                         Text) with button 2 NewDir \0
   NewDir = strip( NewDir)
   if button = \1 & NewDir <> '' & NewDir <> CurStartupDir then
      WriteConfigKey( KeyPath, NewDir)
      'rx startupdir.erx' NewDir
   endif

; ---------------------------------------------------------------------------
; Erasetemp erases a file quietly (no "File not found" message) on both DOS
; and OS/2.  Thanks to Larry Margolis.  Returns 0 if successful erase, or
; the error code (if on DOS) which will usually be 2 for 'file not found'.
defproc EraseTemp( Filename)
   call DeleteFile( Filename)
   return

; ---------------------------------------------------------------------------
; The same as EraseTemp. The statement name 'delete' can't be used as
; procedure name.
defproc DeleteFile( Filename)
   Filename = Filename\0
   rc = dynalink32( 'DOSCALLS',          -- dynamic link library name
                    '#259',              -- ordinal value for Dos32Delete
                    address( Filename))
   return

; ---------------------------------------------------------------------------
defc Delete
   parse arg File1
   File1 = strip( File1)
   if leftstr( File1, 1) = '"' then
      parse value File1 with '"'File1'"'
   endif
   if File1 = '' then
   -- Check if exists
   elseif Exist( File1) then
      -- Confirm delete if Ctrl isn't pressed
      if CtrlIsDown() then
         rc = 0
      else
         if AskOkCancel( 'Delete file "'File1'"?'\n\n ||
            '('CTRL_KEY__MSG' to bypass confirmation.)') then
            rc = 0
         else
            rc = -1
         endif
      endif
      if not rc then
         -- Delete
         call DeleteFile( File1)
         delrc = rc
         -- Check if deleted
         if Exist( File1) then
            'SayError Error: File "'File1'" not deleted.'
         else
            'SayError File "'File1'" deleted.'
         endif
         rc = delrc
      endif
   else
      'SayError Error: File "'File1'" doesn''t exist.'
      rc = 2
   endif

; ---------------------------------------------------------------------------
; Maybe overwrites NewFilename.
defproc CopyFile( Filename, NewFilename)
   Filename    = NepmdQueryFullName( Filename)
   NewFilename = NepmdQueryFullName( NewFilename)

   if Filename = NewFilename then
      rc = 2  -- ERROR_FILE_NOT_FOUND
      return
   endif

   Filename    = Filename\0
   NewFilename = NewFilename\0
   Option = 1  -- Overwrite if exists
   rc = dynalink32( 'DOSCALLS',           -- dynamic link library name
                    '#258',               -- ordinal value for Dos32Copy
                    address( Filename)    ||
                    address( NewFilename) ||
                    atol( Option))
   --dprintf( 'CopyFile: rc = 'rc', 'strip( Filename, 't', \0)' -> 'strip( NewFilename, 't', \0))
   return

; ---------------------------------------------------------------------------
defc CopyFile
   do once = 1 to 1
      parse arg args
      rest = args
      rest = strip( rest)
      if leftstr( rest, 1) = '"' then
         parse value rest with '"'Filename'"' rest
      else
         parse value rest with Filename rest
      endif
      rest = strip( rest)
      if leftstr( rest, 1) = '"' then
         parse value rest with '"'NewFilename'"' rest
      else
         parse value rest with NewFilename rest
      endif

      -- Check if NewFilename doesn't exist or overlay is selected and that
      -- NewFilename is different from Filename.
      do forever
         ChangedFilename = SaveAsCheckExist( NewFilename)
         if rc < 0 then
            leave
         elseif rc = 0 then
            -- nop
         elseif rc = 1 then
            NewFilename = ChangedFilename
         endif
         -- Ensure that NewFilename is different from Filename
         if Filename <> NewFilename then
            leave
         endif
      enddo
      if rc < 0 then
         leave
      endif

      -- Copy file
      Longname = NepmdQueryStringEa( Filename, '.LONGNAME')
      if Longname <> '' then
         UniqueLongname = MakeUniqueLongname( NewFilename, Longname)
      endif
      call CopyFile( Filename, NewFilename)
      if rc then
         'SayError Error on copying file from 'Filename' to 'NewFilename', rc = 'rc'.'
      else
         if GetFileSpec( 'N', Filename) <> GetFileSpec( 'N', NewFilename) then
            if Longname <> '' then
               NepmdWriteStringEa( NewFilename, '.LONGNAME', UniqueLongname)
            endif
         endif
         'SayError File copied to 'NewFilename'.'
      endif
   enddo

; ---------------------------------------------------------------------------
; The original DosMove works only on the same drive. MoveFile handles also
; the move to a different drive.
defproc DosMove( OldFile, NewFile)
   call MoveFile( OldFile, NewFile)
   return

; ---------------------------------------------------------------------------
; Maybe overwrites NewFilename instead of returning ERROR_ACCESS_DENIED.
defproc MoveFile( Filename, NewFilename)
   Filename    = NepmdQueryFullName( Filename)
   NewFilename = NepmdQueryFullName( NewFilename)

   if Filename = NewFilename then
      rc = 2  -- ERROR_FILE_NOT_FOUND
      return
   endif

   if Exist( NewFilename) then
      call DeleteFile( NewFilename)
   endif

   Drive    = GetFileSpec( 'D',  Filename)
   NewDrive = GetFileSpec( 'D',  NewFilename)

   if Drive <> NewDrive | Drive = '\\' | NewDrive = '\\' then
      -- Different drive: use Copy and Delete
      Basename = GetFileSpec( 'B', Filename)

      -- Temporary EA file
      EaFile = GetTmpPath()''Basename'.ea'

      -- Save FileInfo and create EaFile
      call SaveFileInfoEa( Filename, FileInfo, EaFile)

      -- Copy and delete
      call CopyFile( Filename, NewFilename)
      CopyRc = rc
      if not rc then
         call DeleteFile( Filename)
      endif

      -- Restore FileInfo and join EaFile
      call RestoreFileInfoEa( NewFilename, FileInfo, EaFile)

      rc = CopyRc
   else
      -- Same drive: can use Move
      -- DosMove works only on the same drive. Otherwise: rc = 17.
      Filename    = Filename\0
      NewFilename = NewFilename\0
      rc = dynalink32( 'DOSCALLS',          -- dynamic link library name
                       '#271',              -- Dos32Move - move a file
                       address( Filename) ||
                       address( NewFilename), 2)
   endif

   return

; ---------------------------------------------------------------------------
defc MoveFile
   do once = 1 to 1
      parse arg args
      rest = args
      rest = strip( rest)
      if leftstr( rest, 1) = '"' then
         parse value rest with '"'Filename'"' rest
      else
         parse value rest with Filename rest
      endif
      rest = strip( rest)
      if leftstr( rest, 1) = '"' then
         parse value rest with '"'NewFilename'"' rest
      else
         parse value rest with NewFilename rest
      endif

      -- Check if NewFilename doesn't exist or if overlay is selected and
      -- that NewFilename is different from Filename.
      do forever
         ChangedFilename = SaveAsCheckExist( NewFilename)
         if rc < 0 then
            leave
         elseif rc = 0 then
            -- nop
         elseif rc = 1 then
            NewFilename = ChangedFilename
         endif
         -- Ensure that NewFilename is different from Filename
         if Filename <> NewFilename then
            leave
         endif
      enddo
      if rc < 0 then
         leave
      endif

      -- Move file
      Longname = NepmdQueryStringEa( Filename, '.LONGNAME')
      if Longname <> '' then
         UniqueLongname = MakeUniqueLongname( NewFilename, Longname)
      endif
      call MoveFile( Filename, NewFilename)
      if rc then
         'SayError Error on moving file from 'Filename' to 'NewFilename', rc = 'rc'.'
      else
         if GetFileSpec( 'N', Filename) <> GetFileSpec( 'N', NewFilename) then
            if Longname <> '' then
               NepmdWriteStringEa( NewFilename, '.LONGNAME', UniqueLongname)
            endif
         endif
         'SayError File moved to 'NewFilename'.'
      endif
   enddo

; ---------------------------------------------------------------------------
; Reads a file from stdin
defc StdFileRead, StdFile_Read
   InputStream = ''
   ReadBuffer = leftstr( '', 512)  -- Must be < MAXCOL / 2
   BytesReadL = atol( 0)
   do forever
      rc = dynalink32( 'DOSCALLS',                    -- dynamic link library name
                       '#281',                        -- ordinal value for Dos32Read
                       atol( 0)                   ||  -- File handle 0 = STDIN
                       address( ReadBuffer)       ||  -- Buffer area
                       atol( length( ReadBuffer)) ||  -- Buffer area length
                       address( BytesReadL), 2)       -- Bytes read
      if rc then
         'SayError DosRead' ERROR__MSG rc
         return
      endif

      ReadLen = ltoa( BytesReadL, 10)
      if not ReadLen then
         -- Insert rest of InputStream as last line
         if InputStream <> '' then
            insertline InputStream, .last + 1
         endif
         leave
      endif

      fFitInLine = (length( InputStream) + ReadLen) <= MAXCOL
      if fFitInLine then
         -- Append ReadBuffer to InputStream
         InputStream = InputStream''leftstr( ReadBuffer, ReadLen)
      else
         -- Split ReadBuffer at \n chars or when MAXCOL reached
         -- Find first \n char
         pNewline = pos( \n, ReadBuffer)
         if pNewline &
            pNewline < ReadLen &
            (length( InputStream) + ReadLen) <= MAXCOL then
            -- Split InputStream at next \n char
            insertline strip( InputStream''leftstr( ReadBuffer, pNewline - 1),
                              'T', \r), .last + 1
            InputStream = substr( ReadBuffer, pNewline + 1,
                                  ReadLen - pNewline)
         else
            -- Split InputStream at MAXCOL
            l2 = MAXCOL - length( InputStream)
            insertline strip( InputStream''leftstr( ReadBuffer, l2),
                              'T', \r), .last + 1
            InputStream = substr( ReadBuffer, l2 + 1, ReadLen - l2)
         endif
      endif

      -- Split InputStream at next \n chars
      -- Find second \n char
      pNewline = pos( \n, InputStream)
      do while pNewline
         insertline strip( leftstr( InputStream, pNewline - 1),
                           'T', \r), .last + 1
         InputStream = substr( InputStream, pNewline + 1)
         -- Find next \n char
         pNewline = pos( \n, InputStream)
      enddo
   enddo

; ---------------------------------------------------------------------------
; Writes a file to stdout
defc StdFileWrite, StdFile_Write
   universal vtemp_filename

   .filename = vtemp_filename
   's'
   -- Execute the OS/2 cmd 'type'
   quietshell 'type '.filename
   call DeleteFile( .filename)
   'q'

; ---------------------------------------------------------------------------
defproc SaveFileInfoEa( Filename, var FileInfo, var EaFile)
   -- Save Eas
   StripEaCmd = 'eautil' EnquoteFileSpec( Filename) EnquoteFileSpec( EaFile) '/s /r'
   quietshell StripEaCmd
   if rc | not Exist( EaFile) then
      -- Create a 0 byte file if no EAs were found.
      -- This allows for replacing EAs later on join.
      call CopyFile( 'nul', EaFile)
   endif

   -- Save FIL_STANDARD FileInfo
   FileInfo = ''
   Pathname = Filename\0
   ResultBuf = copies( \0, 30)
   rcx = dynalink32( 'DOSCALLS',      -- dynamic link library name
                     '#223',          -- ordinal value for Dos32QueryPathInfo
                     address( Pathname)         ||  -- pathname to be queried
                     atol( 1)                   ||  -- PathInfoLevel
                     address( ResultBuf)        ||  -- buffer where info is to be returned
                     atol( length( ResultBuf)))     -- size of buffer
   --dprintf('Dos32QueryPathInfo returned rc = 'rcx)
   if not rcx then
      FileInfo = ResultBuf
   endif

   -- Calling this has often caused a system trap for large files in Encode.
   -- This is a try to work around that:
   'Sleep '100

   rc = rcx
   return

; ---------------------------------------------------------------------------
defproc RestoreFileInfoEa( Filename, FileInfo, EaFile)
   -- Set FIL_STANDARD FileInfo
   if FileInfo <> '' then
      Pathname = Filename\0
      PathInfo = FileInfo
      rcx = dynalink32( 'DOSCALLS',      -- dynamic link library name
                        '#219',          -- ordinal value for Dos32SetPathInfo
                        address( Pathname)         ||  -- pathname to be altered
                        atol( 1)                   ||  -- PathInfoLevel
                        address( PathInfo)         ||  -- buffer where info is stored
                        atol( length( PathInfo))   ||  -- size of buffer
                        atol( 0))                      -- write option
      --dprintf('Dos32SetPathInfo returned rc = 'rcx)
   endif

   -- Join EAs
   if Exist( EaFile) then
      JoinEaCmd  = 'eautil' EnquoteFileSpec( Filename) EnquoteFileSpec( EaFile) '/j /o'
      quietshell JoinEaCmd
      if not rc then
         call DeleteFile( EaFile)
      endif
   endif

   rc = rcx
   return

; ---------------------------------------------------------------------------
; Used in INFOLINE.E.
defproc GetFileDateHex( Filename)
   Filename = Filename\0
   ResultBuf = copies( \0, 30)
   FileDateHex = ''
   rc = dynalink32( 'DOSCALLS',      -- dynamic link library name
                    '#223',          -- ordinal value for DOS32QueryPathInfo
                    address( Filename)         ||  -- pathname to be queried
                    atol( 1)                   ||  -- PathInfoLevel
                    address( ResultBuf)        ||  -- buffer where info is to be returned
                    atol( length( ResultBuf)))      -- size of buffer
   if rc = 0 then
      FileDateHex = ltoa( substr( ResultBuf, 9, 4), 16)
      -- The return value can be compared with ltoa( substr( .fileinfo, 9, 4), 16)
      -- to determine if the file's timestamp has changed on disk since defload.
      -- (.fileinfo is set at defload.)
   endif
   return FileDateHex

; ---------------------------------------------------------------------------
; For compatibilty.
defproc Get_File_Date( Filename)
   return GetFileDateHex( Filename)

; ---------------------------------------------------------------------------
; Used in INFOLINE.E and BACKUP.E.
defproc FileDateHexToIsoDate( hexstr)
   -- add leading zero if length < 8
   hexstr = rightstr( hexstr, 8, 0)

   date = hex2dec( substr( hexstr, 5, 4))
   year = date % 512
   date = date // 512
   month = date % 32
   day = date // 32 % 1  -- %1 to drop fraction
   date = year + 1980'-'rightstr( month, 2, 0)'-'rightstr( day, 2, 0)  -- ISO date yyyy-mm-dd

   time = hex2dec( substr( hexstr, 1, 4))
   hour = time % 2048
   time = time // 2048
   min = time % 32
   sec = time // 32 * 2 % 1  -- %1 to drop fraction
   time = rightstr( hour, 2, 0)':'rightstr( min, 2, 0)':'rightstr( sec, 2, 0)  -- ISO time hh:mm:ss

   return date time

; ---------------------------------------------------------------------------
; Works with files and dirs. Returns 1 if file/dir exists, otherwise 0.
defproc Exist( Filename)
   SavedRc = rc
   call QueryFileMode( Filename)
   if rc = 0 then
      fExist = 1
   else
      fExist = 0
   endif
   rc = SavedRc
   return fExist

; ---------------------------------------------------------------------------
; Attrib is returned.
defproc QueryFileMode( Filename)
   Filename = Filename\0
   Attrib = copies( \0, 24)  -- allocate 24 bytes for a FileStatus3 structure
   rc = dynalink32( 'DOSCALLS',             -- dynamic link library name
                    '#223',                 -- ordinal value for Dos32QueryPathInfo
                    address( Filename)  ||  -- Pointer to path name
                    atol( 1)            ||  -- PathInfoLevel 1
                    address( Attrib)    ||  -- Pointer to info buffer
                    atol( 24), 2)           -- Buffer Size
   Attrib = ltoa( rightstr( Attrib, 4), 10)
   return Attrib

; ---------------------------------------------------------------------------
; Old def, Attrib must be specified as arg( 2), rc is returned.
defproc QFileMode( Filename, var Attrib)
   Attrib = QueryFileMode( Filename)
   return rc

; ---------------------------------------------------------------------------
; arg( 1) = drive, e.g.: X:
; Returns e.g.: HPFS | CDFS | JFS | FAT | LAN | RAMFS | ISOFS | HPFS386 | FAT32 | NDFS32 | NTFS
defproc QueryFileSys( Drive)
   dev_name = Drive\0
   ordinal = 0
   infobuf = substr( '', 1, 255)
   infobuflen = atol( length( infobuf))
   FSAinfolevel = 1
   rc = dynalink32( 'DOSCALLS',              -- dynamic link library name
                    '#277',                  -- ordinal value for Dos32QueryFSAttach
                    address( dev_name)  ||   -- device name
                    atol( ordinal)      ||   --
                    atol( FSAinfolevel) ||   -- info level requested
                    address( infobuf)   ||   -- string offset
                    address( infobuflen), 2) -- length of buffer
   if rc then
      return
   else
      -- For FSAinfolevel 1:
      iType = itoa( substr( infobuf, 1, 2), 10)
           -- 1=resident char. dev.; 2=pseudochar dev.; 3=local drive; 4=remote drive
      cbName = itoa( substr( infobuf, 3, 2), 10)
      cbFSDName = itoa( substr( infobuf, 5, 2), 10)
      cbFSAData = itoa( substr( infobuf, 7, 2), 10)
      parse value substr( infobuf, 9) with szName \0 szFSDName \0 rgFSAData
      rgFSAData = leftstr( rgFSAData, cbFSAData)
      --insertline 'dev='dev_name 'iType=' iType 'cbName='cbName 'cbFSDName='cbFSDName 'cbFSAData='cbFSAData, .last+1
      --insertline '      szName="'szName'" szFSDName="'szfsdname'" rgFSAData="'rgFSAData'"', .last+1
      return szFSDName
   endif

; ---------------------------------------------------------------------------
defproc GetTmpDir
   do once = 1 to 1
      TmpDir = Get_Env( 'TMP')
      if TmpDir <> '' then
         leave
      endif
      TmpDir =  Get_Env( 'TEMP')
      if TmpDir <> '' then
         leave
      endif
   enddo
   return TmpDir

defproc GetTmpPath
   return strip( GetTmpDir(), 'T', '\')'\'  -- append a backslash

; ---------------------------------------------------------------------------
defproc GetTempDir
   return GetTmpDir()

defproc GetTempPath
   return GetTmpPath()

; ---------------------------------------------------------------------------
; Accepts a relative dir name. The parent dir must exist.
defproc MakeDirectory( Dirname)
   DirName = NepmdQueryFullName( DirName)
   Dirname = Dirname\0
   peaop2 = copies( \0, 4)
   rc = dynalink32( 'DOSCALLS',           -- dynamic link library name
                    '#270',               -- ordinal value for Dos32CreateDir
                    address( Dirname) ||  -- device name
                    peaop2)
   return

; ---------------------------------------------------------------------------
; Accepts a relative tree name.
defproc MakeTree( DirName)

   -- Remove trailing backslash. Keep it for root dirs to return a better rc.
   do once = 1 to 1
      if rightstr( DirName, 1) <> '\' then
         leave
      elseif DirName = '\' then
         leave
      elseif rightstr( DirName, 2) = ':\' then
         leave
      endif
      DirName = strip( DirName, 't', '\')
   enddo

   DirName = NepmdQueryFullName( DirName)

   rc = 0
   p = pos( '\', DirName)
   if p = 0 then
      rc = 13 -- ERROR_INVALID_DATA
      return
   endif
   -- Ignore first backslash
   next = substr( DirName, 1, p - 1)
   rest = substr( DirName, p)

   -- Loop through dir segments. next is the full path
   -- without backslash and rest has a leading backslash.
   fbreak = 0
   do forever
      p = pos( '\', rest, 2)
      if p = 0 then
         next = next''rest
         rest = ''
         fbreak = 1
      else
         next = next''substr( rest, 1, p - 1)
         rest = substr( rest, p)
      endif

      SubDirname = next\0
      peaop2 = copies( \0, 4)
      rcx = dynalink32( 'DOSCALLS',             -- dynamic link library name
                        '#270',                 -- ordinal value for Dos32CreateDir
                        address( SubDirname) || -- device name
                        peaop2)

      -- rcx = 5 (ERROR_ACCESS_DENIED) is returned if dir already exists
      if not wordpos( rcx, '0 5') then
         rc = rcx
         leave
      endif
      if fbreak then
         leave
      endif
   enddo

   return

; ---------------------------------------------------------------------------
; Gets the real case of a file system object. Returns the input if it doesn't
; exist. Doesn't change rc.
; DosFind* for a filename or dirname returns the correct case, but only for
; the last segment. To get the correct case for the entire string, it must
; be processed segment-wise.
defproc GetRealCase( FileMask)
   fFile = 0
   fDir  = 0
   savedrc = rc

   do once = 1 to 1
      -- Process only full file specs
      if substr( FileMask, 2, 1) <> ':' then
         leave
      endif

      -- Check if dir or file to call the correct proc for the last segment
      if NepmdFileExists( FileMask) then
         fFile = 1
      elseif NepmdDirExists( FileMask) then
         fDir  = 1
      -- Handle drive only
      elseif rightstr( FileMask, 1) = ':' then
         FileMask = upcase( FileMask)
         leave
      else
         -- Process only existing file specs
         leave
      endif

      rest = FileMask
      NewFileMask = ''
      do while rest <> ''
         parse value rest with next'\'rest
         if next = '' then
            iterate
         endif

         -- Drive
         if rightstr( next, 1) = ':' then
            NewFileMask = NewFileMask''upcase( next)
            -- For root dir: append '\' here, because no more segemnt exists
            if rightstr( FileMask, 2) = ':\' then
               NewFileMask = NewFileMask'\'
            endif
            iterate
         endif

         if rest = '' then
            -- Last segment
            if fFile then
               NewFileMask = NewFileMask'\'next
               Handle      = ''    -- always create a new handle!
               FoundName   = ''
               if NepmdGetNextFile( NewFileMask, Handle, FoundName) then
                  NepmdGetNextClose( Handle)
                  NewFileMask = FoundName
               endif
            elseif fDir then
               NewFileMask = NewFileMask'\'next
               Handle      = ''    -- always create a new handle!
               FoundName   = ''
               if NepmdGetNextDir( NewFileMask, Handle, FoundName) then
                  NepmdGetNextClose( Handle)
                  NewFileMask = FoundName
               endif
            endif

         else
            -- Other segments
            NewFileMask = NewFileMask'\'next
            Handle      = ''    -- always create a new handle!
            FoundName   = ''
            if NepmdGetNextDir( NewFileMask, Handle, FoundName) then
               NepmdGetNextClose( Handle)
               NewFileMask = FoundName
            endif
         endif

      enddo

      FileMask = NewFileMask
   enddo
   rc = savedrc

   return FileMask

; ---------------------------------------------------------------------------
; Returns a substring of FileSpec, depending on Option. Following Options
; exist:
;    D = drive (or server for a UNC name or protocol id for a URL)
;    P = path (with trailing backslash or slash)
;    N = name
;    B = basename (without trailing dot and extension)
;    E = extension (without leading dot)
; Any combination of these single options may be specified, e.g.
; OptionStr = 'DPB' returns the full filename without dot and extension.
; Other chars than D, P, N, B or E are not allowed, e.g. it's not possible
; to specify 'B.E' instead of 'N'.
; Only the first letter of Option is read.
defproc GetFileSpec( OptionStr, FileSpec)
   do once = 1 to 1
      -- Init vars
      OutStr   = ''
      Drive    = ''
      Path     = ''
      Name     = ''
      Basename = ''
      Ext      = ''

      -- Check args
      if FileSpec = '' then
         leave
      endif
      if OptionStr = '' then
         leave
      endif
      OptionStr = upcase( OptionStr)
      if verify( OptionStr, 'DPNBE') then
         leave
      endif

      fSlashFirst = 0
      -- Find the first slash char
      p = verify( FileSpec, '/\', 'M')
      if p then
         ThisSlash = substr( FileSpec, p, 1)
         if ThisSlash = '/' then
            fSlashFirst = 1
         endif
      endif

      -- Convert all slash chars to backslash
      Rest = translate( FileSpec, '\', '/')

      fUnc = (leftstr( Rest, 2) == '\\')
      fUrl = (pos( ':\\', Rest) <> 0)

      -- Parse Drive
      if fUnc then
         p = pos( '\', Rest, 3)  -- skip the first '\\'
         if p then
            Drive = leftstr( Rest, p - 1)
            Rest  = substr( Rest, p)
         else
            Drive = Rest
            Rest  = ''
         endif
      elseif fUrl then
         p = pos( ':\\\', Rest)
         if p then
            Drive = leftstr( Rest, p + 3)  -- incl. ':\\\'
            Rest  = substr( Rest, p + 3 + 1)
         else
            p = pos( ':\\', Rest)
            if p then
               Drive = leftstr( Rest, p + 2)  -- incl. ':\\'
               Rest  = substr( Rest, p + 2 + 1)
            endif
         endif
      else
         p = pos( ':\', Rest)
         if p then
            Drive = leftstr( Rest, p)  -- incl. ':'
            Rest  = substr( Rest, p + 1)
         endif
      endif

      -- Parse Path and Name
      p1 = pos( '\', Rest)
      if p1 then
         p2 = lastpos( '\', Rest)
         if p2 > p1 then
            Path = leftstr( Rest, p2)
            Name = substr( Rest, p2 + 1)
         else
            Path = leftstr( Rest, p1)
            Name = substr( Rest, p1 + 1)
         endif
      else
         Name = Rest
      endif

      -- Parse Basename and Ext
      p = lastpos( '.', Name)
      if p then
         Basename = leftstr( Name, p - 1)
         Ext      = substr( Name, p + 1)
      else
         Basename = Name
      endif

      do p = 1 to length( OptionStr)
         Option = substr( OptionStr, p, 1)
         if Option = 'D' then
            OutStr = OutStr''Drive
         elseif Option = 'P' then
            OutStr = OutStr''Path
         elseif Option = 'N' then
            OutStr = OutStr''Name
         elseif Option = 'B' then
            OutStr = OutStr''Basename
         elseif Option = 'E' then
            OutStr = OutStr''Ext
         endif
      enddo

      if fSlashFirst then
         -- Convert all backslash chars to slash
         OutStr = translate( OutStr, '/', '\')
      endif
   enddo
   return OutStr

; ---------------------------------------------------------------------------
; Avoids double dirs. CurDir is used for '.' in SearchPath.
defproc NormalizeSearchPath( SearchPath)
   CurDir = arg( 2)
   if CurDir = '' then
      CurDir = NepmdQueryFullname( .filename'\..')
   endif

   Rest = SearchPath
   SearchPath = ''
   do while Rest <> ''
      parse value Rest with ThisDir';'Rest
      if ThisDir = '' then
         iterate
      endif

      -- ThisDir may have a recurse spec appended, so it can't be chacked
      -- if it exists.

      -- Expand env vars for single dirs her here to check for double
      -- entries and to recognize unset env vars.
      DirEnv = Get_Env( ThisDir)
      if DirEnv <> '' then
         ThisDir = DirEnv
      else
         fSetRc = 1  -- Without this, rc is not set by ExpandEnvVars
         ThisDir = ExpandEnvVars( ThisDir, fSetRc)
         if rc then  -- 204, if an env var is unset or empty
            -- Avoid searching an unwanted dir or tree, if an env var is
            -- unset, e.g. %TOOLKIT_DIR%\H would expand to \H
            iterate
         endif
      endif

      -- Expand '.' to CurDir
      if ThisDir = '.' then
         ThisDir = CurDir
      endif

      -- Make relative dir specs work relative to CurDir
      SavedDir = directory()
      call Directory2( CurDir)
      ThisDir = NepmdQueryFullname( ThisDir)
      call Directory2( SavedDir)

      -- Check if ThisDir is already added
      if pos( ';'ThisDir';', ';'SearchPath) then
         iterate
      endif

      -- Add ThisDir
      if ThisDir <> '' then
         SearchPath = SearchPath''ThisDir';'
      endif
   enddo

   return SearchPath

; ---------------------------------------------------------------------------
; Checks DirSpec for optional trailing numbers and a trailing exclamation
; mark. It returns Dir and RecurseLevel.
; o  A trailing '!' issues a recursive search in that directory and all its
;    subdirectories.
; o  To limit the recursion, a number may be appended after the '!'.
;    '!0' means no search in subdirectories. A good value is '!3'.
; o  '!F' means fully recursive, the same as '!'.
; o  '!V' means variable. The recurse depth depends on the amount of subpaths
;    to avoid searching a whole volume.
defproc ParseDirRecurseLevel( DirSpec, var Dir, var RecurseLevel)
   -- Init output vars
   Dir = DirSpec
   RecurseLevel = 0
   do once = 1 to 1
      -- Check if a '!' is at the end, optionally with an appendix
      TestDir = strip( translate( DirSpec, copies( ' ', 14), '0123456789VFvf'))
      if rightstr( TestDir, 1) <> '!' then
         leave
      endif
      -- Check for an appendix after '!'
      pExMark = lastpos( '!', DirSpec)
      -- No appendix
      if pExMark = length( DirSpec) then
         RecurseLevel = -1  -- fully recursive
         Dir = leftstr( DirSpec, pExMark - 1)
      -- Appendix
      else
         -- Number of recursive search levels is appended after the '!'
         RecurseLevel = substr( DirSpec, pExMark + 1)
         if upcase( RecurseLevel) = 'F' then
            RecurseLevel = -1  -- fully recursive
         elseif upcase( RecurseLevel) = 'V' then
            -- Number of start sublevels = number of backslash chars.
            -- This allows for specifying recursive search without searching
            -- a whole volume when starting near the root.
            NumSubLevels = count( '\', DirSpec)  -- variable recursion
            if NumSubLevels >= 3 then
               RecurseLevel = -1
            elseif NumSubLevels >= 2 then
               RecurseLevel = 3
            elseif NumSubLevels >= 1 then
               RecurseLevel = 2
            endif
         endif
         Dir = leftstr( DirSpec, pExMark - 1)
      endif
   enddo

   return

; ---------------------------------------------------------------------------
; Syntax: FindFileInList( File, DirList [, SubDirList])
; Searches file or dir in a list of dir specifications, dirs separated by
; ";".
; o  A trailing '!' issues a recursive search in that directory and all its
;    subdirectories.
; o  To limit the recursion, a number may be appended after the '!'.
;    '!0' means no search in subdirectories. A good value is '!3'.
; o  '!F' means fully recursive, the same as '!'.
; o  '!V' means variable. The recurse depth depends on the amount of subpaths
;    to avoid searching a whole volume.
; o  If both the current part of DirList and the current part of SubDirList
;    have recurse levels specified, the one for DirList has precedence.
;    Its recurse level value is then applied to all its combinations with a
;    subdir.
; o  It's normally not useful to search in subdirs of a search path. To
;    avoid this, use multiple calls of FindFileInList, without and with
;    subpath arg.
; o  This can also be used instead of findfile to avoid a search in the
;    current dir.
; Examples:
;    FullExName = FindFileInList( 'mouse.ex', Get_Env( 'EPMEXPATH'))
;    FullExName = FindFileInList( 'mouse.ex', '%EPMEXPATH%')
;    FullExName = FindFileInList( 'mouse.ex', '..;EPMEXPATH')
;    FullExName = FindFileInList( 'mouse.ex', 'EPMEXPATH;PATH')
;    FullExName = FindFileInList( 'mouse.ex', 'EPMEXPATH', '\.;\..;\..\ex')
;    FullExName = FindFileInList( 'mouse.ex', 'EPMEXPATH', '.;..;..\ex')
;    FullEName  = FindFileInList( 'mouse.e', .filename'\..', '.!;..!;..\..!')
;    FullEName  = FindFileInList( 'small.e', .filename'\..', '.!;..!;..\..!')
;    FullEName  = FindFileInList( 'small.e', .filename'\..', '.!3;..!3;..\..!3')
;    FullEName  = FindFileInList( 'small.e', .filename'\..', '.!;..!3')
;    FullName   = FindFileInList( 'stdlib.h', 'G:\dev\toolkt45\H;G:\dev\toolkt45\IDL;')  (not found)
;    FullName   = FindFileInList( 'stdlib.h', 'G:\dev\toolkt45\H!;G:\dev\toolkt45\IDL;')
; Returns '' if File not found in DirList, otherwise fully qualified name.
defproc FindFileInList( File, DirList)
   do once = 1 to 1
      SubDirList = arg( 3)
      Cmd        = arg( 4)
      rc = -1
      --dprintf( 'FindFileInList( 'File', 'DirList', 'SubDirList', 'Cmd')')
      FullName = ''

      -- First expand '?:' to make other wildcard processing easier
      if leftstr( File, 2) = '?:' then
         -- Replace ?: with bootdrive
         BootDrive = NepmdQuerySysInfo( 'BOOTDRIVE')
         File = ChangeStr( '?:', File, BootDrive)
      endif

      WildcardPos = verify( File, '*?', 'M')

      -- Check if not fully qualified
      if not WildcardPos then
         if substr( File, 2, 2) = ':\' |     -- fully qualified local spec
            substr( File, 1, 2) = '\\' then  -- fully qualified UNC spec
            FullName = File
            rc = 0
            leave
         endif
      endif

      -- DirList may contain env vars. Expand them per ThisDir instead of
      -- at once to not extemd 1600 chars.
      rest = DirList
      do while rest <> ''
         fFound = 0
         parse value rest with ThisDir';'rest
         if ThisDir = '' then
            iterate
         endif
         -- Try to expand envvars
         DirEnv = Get_Env( ThisDir)
         if DirEnv <> '' then
            ThisDir = DirEnv
         else
            -- The 1 flag makes ExpandEnvVars set rc. It is set to 204, if an
            -- env var is unset or empty.
            fSetRc = 1
            ThisDir = ExpandEnvVars( ThisDir, fSetRc)
            if rc then
               -- Avoid searching in an unwanted dir or tree, if an env var is
               -- unset, e.g. %TOOLKiT_DIR%\H would expand to \H
               iterate
            endif
         endif

         -- After expanding wildcards, ThisDir itself may consist of multiple dirs.
         -- Parse ThisDir here again.
         restenv = ThisDir
         do while restenv <> ''
            parse value restenv with ThisDir';'restenv
            if ThisDir = '' then
               iterate
            endif

            call ParseDirRecurseLevel( ThisDir, ThisDir, RecurseLevel)

            SearchDirList = ''

            -- Search in subdirs of Dir, if specified
            if SubDirList = '' then
               SearchDirList = ThisDir
            else
               restsub = SubDirList
               do while restsub <> ''
                  parse value restsub with ThisSub';'restsub
                  if ThisSub = '' then
                     iterate
                  endif
                  -- Append ThisDir'\'ThisSub to SearchDirList
                  SearchDirList = SearchDirList''strip( ThisDir, 'T', '\')'\'strip( ThisSub, 'L', '\')';'
               enddo
            endif

            restsearch = SearchDirList
            do while restsearch <> ''

               fFound = 0
               parse value restsearch with ThisSearchDir';'restsearch
               if ThisSearchDir = '' then
                  iterate
               endif

               -- Save recurse level from above
               SavedRecurseLevel = RecurseLevel

               call ParseDirRecurseLevel( ThisSearchDir, ThisSearchDir, RecurseLevel)

               -- Restore saved recurse level. This gives the recurse level
               -- specified in a DirList segment priority. Then
               -- ParseDirRecurseLevel is just used to strip the '!' part.
               if SavedRecurseLevel <> 0 then
                  RecurseLevel = SavedRecurseLevel
               endif

               ThisSearchDir = NepmdQueryFullname( ThisSearchDir)
               --dprintf( 'ThisSearchDir = 'ThisSearchDir', RecurseLevel = 'RecurseLevel)
               if not NepmdDirExists( ThisSearchDir) then
                  iterate
               endif

               -- If File has wildcards, call SearchFileInTree
               if WildcardPos then
                  -- This opens a listbox with found files
                  call SearchFileInTree( File, ThisSearchDir, RecurseLevel, Cmd)
                  return
               else
                  if RecurseLevel <> 0 then
                     --dprintf( 'NepmdSearchTree( 'File', 'ThisSearchDir', 'RecurseLevel')')
                     FullName = NepmdSearchTree( File, ThisSearchDir, RecurseLevel)
                  else
                     next = strip( ThisSearchDir, 'T', '\')'\'File
                     if Exist( next) then  -- find files and dirs
                        FullName = NepmdQueryFullname( next)
                     endif
                  endif
                  if FullName <> '' then
                     leave
                  endif
               endif

            enddo  -- restsearch of SearchDirList
            if FullName <> '' then
               leave
            endif
         enddo  -- restenv of ThisDir (expandded env vars may contain multiple dirs)
         if FullName <> '' then
            leave
         endif
      enddo  -- rest of DirList
   enddo  -- once

   return FullName

; ---------------------------------------------------------------------------
defproc SearchFileInTree( FileMask, Dir)
   do once = 1 to 1
      rc = 2  -- File not found
      RecurseLevel = arg( 3)
      Cmd          = arg( 4)
      --dprintf( 'SearchFileInTree( 'FileMask', 'Dir', 'RecurseLevel', 'Cmd')')

      call ParseDirRecurseLevel( Dir, Dir, RecurseLevel)

      getfileid startfid
      tmpfid = LoadHiddenFile( '.found')     -- for ListBox
      tmp2fid = LoadHiddenFile( '.found2')   -- for Open all
      call SearchFileInTree2( FileMask, Dir, tmpfid, tmp2fid, RecurseLevel, Cmd)
   enddo

   if not rc then
      display -14
      call SearchShowFoundFiles( tmpfid, tmp2fid, startfid)
      display 14
   endif

   -- Ensure that all hidden files are closed
   call CloseHiddenFile( tmpfid)
   call CloseHiddenFile( tmp2fid)

   return

defc SearchFileInTree
   parse arg FileMask Dir
   call SearchFileInTree( FileMask, Dir)

; ---------------------------------------------------------------------------
; This sets rc = 0 if an entry was added to tmpfid.
defproc SearchFileInTree2( ThisMask, ThisDir, tmpfid, tmp2fid, RecurseLevel)
   do once = 1 to 1
      Cmd = arg( 6)

      -- Process files
      if substr( ThisMask, 2, 2) = ':\'  then
         FileMask = ThisMask
      elseif leftstr( ThisMask, 2) = '\\'  then
         FileMask = ThisMask
      elseif leftstr( ThisMask, 1) = '\' then
         FileMask = GetFileSpec( 'D', ThisDir)ThisMask
      else
         FileMask = strip( ThisDir, 'T', '\')'\'ThisMask
      endif

      Handle   = ''  -- always create a new handle
      FullName = ''
      savedrc = rc
      --dprintf( 'SearchFileInTree2: before NepmdGetNextFile: FileMask = 'FileMask)
      do while NepmdGetNextFile( FileMask, Handle, FullName)
         --dprintf( 'SearchFileInTree2: NepmdGetNextFile( 'FileMask', Handle, 'FullName')')
         rc = savedrc

         if Cmd = '' then
            LineStr = FullName
         else
            LineStr = FullName "'"Cmd"'"
         endif

         call AppendHiddenFile( tmpfid, LineStr)
         call AppendHiddenFile( tmp2fid, LineStr)

      enddo
      rc = savedrc

      -- Set rc
      if tmpfid.last > 1 then
         rc = 0
      endif

      NextRecurseLevel = RecurseLevel
      if RecurseLevel < 0 then
         -- Keep it, -1 means fully recursive
      elseif RecurseLevel > 0 then
         NextRecurseLevel = RecurseLevel - 1
      endif

      if RecurseLevel <> 0 then
         -- Process subdirs
         DirMask     = strip( ThisDir, 'T', '\')'\*'
         Handle      = ''  -- always create a new handle
         NextDirname = ''
         savedrc = rc
         do while NepmdGetNextDir( DirMask, Handle, NextDirname)
            -- Recursive call
            rc = savedrc
            call SearchFileInTree2( ThisMask, NextDirname, tmpfid, tmp2fid,
               NextRecurseLevel, Cmd)
            savedrc = rc
         enddo
         rc = savedrc
      endif
   enddo

   return

; ---------------------------------------------------------------------------
defproc SearchShowFoundFiles( tmpfid, tmp2fid, startfid)

   do once = 1 to 1
      --dprintf( 'SearchShowFoundFiles( 'tmpfid', 'tmp2fid', 'startfid')')
      if not validatefileid( tmpfid) then
         leave
      endif
      if tmpfid.last > 1 then
         activatefile tmpfid
         nFiles = .last - 1

         -- Open ListBox
         Title = 'Select a file'copies( ' ', 20)
         Text = nFiles 'file(s) found.'
         'SayError 'Text
         Text = CTRL_KEY__MSG' to open a new window'
         DefaultItem = 1
         Buttons = '/~Open/Open ~all/'CANCEL__MSG
         DefaultButton = 1
         HelpId = 0

         if ListBox_Buffer_From_File( startfid, bufhndl, noflines, usedsize) then
            leave
         endif
         ret = ListBox( Title,
                        \0 || atol( usedsize) || atoi( 32) || atoi( bufhndl),
                        Buttons,
                        0, 0,   -- top, left,
                        Min( 20, Max( 4, noflines - 1)), 80,  -- height, width
                        GethWndC( APP_HANDLE) ||
                        atoi( DefaultItem)    ||
                        atoi( DefaultButton)  ||
                        atoi( HelpId)         ||
                        Text\0)
         call buffer( FREEBUF, bufhndl)

         -- Parse return string
         parse value ret with Button 2 Select \0
         Button = asc( Button)

         if Button < 1 | Button > 2 then  -- Cancel
            rc = 31  -- Cancel
            leave
         endif

         -- Set number and list of files to load
         if Button = 1 then  -- load one file
            EditArgs = Select
         elseif Button = 2 then  -- load all files
            if validatefileid( tmp2fid) then
               -- Save tmp2fid as ListFile
               ListFile = GetTmpDir()'\nepmd\tmp.epm'
               call SaveCloseHiddenFile( tmp2fid, ListFile, startfid)
               EditArgs = '@'ListFile
            endif
         endif
         rc = 0
         call DialogLoadFile( EditArgs)
      else
         rc = 2
         'SayError No file found.'
      endif
   enddo
   return

; ---------------------------------------------------------------------------
; Checks for signature, returns 1 if found in SigList, 0 if not, '' if error.
; Sets rc. SigList is a space-separated list of signatures.
defproc CheckSig( File, SigList)
   Result = 0
   MaxLen = 0
   do w = 1 to words( SigList)
      Sig = word( SigList, w)
      MaxLen = Max( MaxLen, length( Sig))
   enddo

   FileSig = NepmdReadFilePart( File, MaxLen)
   if rc <> 0 then
      return ''
   endif

   do w = 1 to words( SigList)
      Sig = word( SigList, w)
      if leftstr( FileSig, length( Sig)) = Sig then
         Result = 1
         leave
      endif
   enddo

   return Result

; ---------------------------------------------------------------------------
; Returns the filename of a line in a .TreeDir or .Tree file.
defproc GetTreeListFilename
   LineStr = textline( .line)
   ExtractFilename = ''
   do once = 1 to 1
      -- Check filename
      if upcase( leftstr( .filename, 5)) <> '.TREE' then
         leave
      endif
      -- Must use positional parsing, because Attr may be empty
      -- Check date and time cols
      if substr( LineStr, 5, 1)''substr( LineStr, 8, 1)''substr( LineStr, 15, 1) ||
         substr( LineStr, 18, 1) <> '--::' then
         leave
      endif
      ExtractFilename = substr( LineStr, 52)
   enddo
   return ExtractFilename

; ---------------------------------------------------------------------------
; Returns 1 if on a line with a tree filename and if a file, 2 if on a <dir>
; line and otherwise 0.
defproc IsATreeListFilename
   LineStr = textline( .line)
   ExtractFilename = ''
   do once = 1 to 1
      -- Check filename
      if upcase( leftstr( .filename, 5)) <> '.TREE' then
         rcx = 0
         leave
      endif
      -- Must use positional parsing, because Attr may be empty
      -- Check date and time cols
      if substr( LineStr, 5, 1)''substr( LineStr, 8, 1)''substr( LineStr, 15, 1) ||
         substr( LineStr, 18, 1) <> '--::' then
         rcx = 0
         leave
      endif
      FilesizeVal = strip( substr( LineStr, 23, 9))
      if IsNum( FileSizeVal) then
         rcx = 1
      elseif upcase( FileSizeVal) = '<DIR>' then
         rcx = 2
      else
         rcx = 0
      endif
   enddo
   return rcx

; ---------------------------------------------------------------------------
defproc IsADirList
   rcx = 0
   -- Check filename
   do once = 1 to 1
      if upcase( leftstr( .filename, 8)) = '.DOS DIR' then
         rcx = 1
      elseif upcase( leftstr( ShellGetLastCmd(), 4)) = 'DIR ' then
         rcx = 1
      elseif not IsAShell() then
         leave
      endif
   enddo
   return rcx

; ---------------------------------------------------------------------------
; Returns the filename of a line in a .DOS dir file.
defproc ParseDirListLine( var fFile, var fDir, var Date, var Time, var Size,
                          var EaSize, var Fileattr, var Filename)
   LineStr = textline( .line)
   fDir  = 0
   fFile = 0
   Date     = ''
   Time     = ''
   Size     = ''
   EaSize   = ''
   Fileattr = ''
   Filename = ''
   do once = 1 to 1
      if not IsADirList() then
         leave
      endif

      -- Parse words to allow any date format
      -- Non-FAT
      ThisDate     = strip( subword( LineStr, 1, 1))
      ThisTime     = strip( subword( LineStr, 2, 1))
      ThisSize     = strip( subword( LineStr, 3, 1))
      ThisEaSize   = strip( subword( LineStr, 4, 1))
      ThisFileattr = strip( subword( LineStr, 5, 1))
      ThisFilename = strip( subword( LineStr, 6))

      if verify( ThisDate, DIR_DATETIME_CHARS) |
         verify( ThisTime, DIR_DATETIME_CHARS) |
         verify( ThisFileattr, DIR_ATTRIB_CHARS) |
         (verify( ThisSize, DIR_SIZE_CHARS) & upcase( ThisSize) <> '<DIR>') then
         -- Try FAT listing
         ThisBasename = strip( subword( LineStr, 1, 1))
         ThisExt      = strip( subword( LineStr, 2, 1))
         if upcase( ThisExt) = '<DIR>' then
            ThisFilename = ThisBasename
            ThisSize     = ThisExt
            ThisExt      = ''
            ThisDate     = strip( subword( LineStr, 3, 1))
            ThisTime     = strip( subword( LineStr, 4, 1))
            ThisFileAttr = ''
         else
            ThisFilename = ThisBasename'.'ThisExt
            ThisSize     = strip( subword( LineStr, 3, 1))
            ThisDate     = strip( subword( LineStr, 4, 1))
            ThisTime     = strip( subword( LineStr, 5, 1))
            ThisFileAttr = ''
         endif
      endif

      -- Check date, time, attrib, size cols
      if verify( ThisDate, DIR_DATETIME_CHARS) then
         leave
      elseif verify( ThisTime, DIR_DATETIME_CHARS) then
         leave
      elseif verify( ThisFileattr, DIR_ATTRIB_CHARS) then
         leave
      elseif verify( ThisSize, DIR_SIZE_CHARS) then
         if upcase( ThisSize) = '<DIR>' then
            fDir = 1
         else
            leave
         endif
      elseif ThisFilename <> '' then
         fFile = 1
      endif

      Date     = ThisDate
      Time     = ThisTime
      Size     = ThisSize
      EaSize   = ThisEaSize
      Fileattr = ThisFileattr
      Filename = ThisFilename
   enddo
   return

; ---------------------------------------------------------------------------
; Returns the filename or dirname of a line in a .DOS dir file.
; The 'Directory of' value is prepended to get a full pathname.
defproc GetDirListFilename
   call ParseDirListLine( fFile, fDir, Date, Time, Size,
                          EaSize, Fileattr, Filename)
   ret = ''
   if pos( ':\', Filename) | pos( ':/', Filename) | pos( '\\', Filename) then
      ret = Filename
   else
      call pSave_Pos( SavedPos)
      DirectoryOfStr = GetDirectoryOfStr()

      -- Find DirectoryOfStr
      display -2
      'xcom l '\1'^:o'DirectoryOfStr''\1'x-r'
      display 2
      FoundLen = getpminfo( EPMINFO_LSLENGTH)

      -- Parse ParentDir
      LineStr = textline( .line)
      ParentDir = strip( substr( LineStr, FoundLen + 1))
      -- Handle 4os2 output that lists not only the name of the dir, but
      -- appends the wildcard segment, if specified:
      lp = lastpos( '\', ParentDir)
      LastSeg = substr( ParentDir, lp)
      if verify( LastSeg, '*?', 'M') then
         -- Strip last segment
         ParentDir = substr( ParentDir, 1, lp - 1)
      endif

      -- Prepend to Filename
      ret = strip( ParentDir, 'T', '\')'\'Filename
      call pRestore_Pos( SavedPos)
   endif
   return ret

; ---------------------------------------------------------------------------
; Returns 1 if on a line with a dir filename and if a file, 2 if on a <dir>
; line and otherwise 0.
defproc IsADirListFilename
   call ParseDirListLine( fFile, fDir, Date, Time, Size,
                          EaSize, Fileattr, Filename)
   if fFile then
      return 1
   elseif fDir then
      return 2
   else
      return 0
   endif

; ---------------------------------------------------------------------------
; Returns 1 if on a line with a dir filename or dirname, otherwise 0.
defproc OnADirListLine
   return( IsADirListFilename() > 0)

; ---------------------------------------------------------------------------
; Deletes file under cursor for a .Tree or a .DOS dir file
defc DeleteListFile
   AltCmd = strip( arg( 1))
   if IsATreeListFilename() then
      pSave_Pos( SavedPos)
      Filename = GetTreeListFilename()
      'Delete' EnquoteFilespec( Filename)
      if not rc then
         if pos( '.TREEBACKUP', upcase( .filename)) = 1 then
            -- Reload current window
            Cmd = substr( .filename, 2)
            Cmd
            'PostMe RestorePos' SavedPos
         endif
      endif
   elseif IsADirListFilename() then
      Filename = GetDirListFilename()
      if Filename <> '' then
         'Delete' EnquoteFilespec( Filename)
      endif
   else
      AltCmd
   endif

; ---------------------------------------------------------------------------
defproc ParseTreeTitleFilename( Args, var Filename, var SortOption)
   -- Strip appended SortOption
   if leftstr( Args, 1) = '"' then
      parse value Args with '"'Filename'"' SortOption
   else
      parse value Args with Filename SortOption
   endif
   Filename   = strip( Filename)
   SortOption = strip( SortOption)

